"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
random_line_sample_image(h, w, scan_length, num_points, discrete=None,
    seed=None)
    Function for random line sampling an image.
random_line_sample_surface(l, w, speed, sample_rate, time, discrete=None,
    seed=None)
    Function for random line sampling a surface.

"""

from __future__ import division

import numpy as np

from magni.imaging.measurements import _util
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['random_line_sample_image', 'random_line_sample_surface']

_min_l = _util.min_l
_min_w = _util.min_w
_min_speed = _util.min_speed
_min_sample_rate = _util.min_sample_rate
_min_time = _util.min_time
_min_scan_length = _util.min_scan_length
_min_num_points = _util.min_num_points


def random_line_sample_image(h, w, scan_length, num_points, discrete=None,
                             seed=None):
    """
    Sample an image using a set of random straight lines.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using a pattern based on a set of random straight lines
    are determined. The `scan_length` determines the length of the path scanned
    whereas `num_points` indicates the number of samples taken on that path. If
    `discrete` is set, it specifies the finite number of equally spaced lines
    from which the scan lines are be chosen at random. For reproducible
    results, the `seed` may be used to specify a fixed seed of the random
    number generator.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.
    discrete : int or None, optional
        The number of equally spaced lines from which the scan lines are chosen
        (the default is None, which implies that no discritisation is used).
    seed : int or None, optional
        The seed used for the random number generator (the defaul is None,
        which implies that the random number generator is not seeded).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis.

    Each of the scanned lines span the entire width of the image with the
    exception of the last line that may only be partially scanned if the
    `scan_length` implies this. The top and bottom lines of the image are
    always included in the scan.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import random_line_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> seed = 6021
    >>> np.set_printoptions(suppress=True)
    >>> random_line_sample_image(h, w, scan_length, num_points, seed=seed)
    array([[ 0.5       ,  0.5       ],
           [ 4.59090909,  0.5       ],
           [ 8.68181818,  0.5       ],
           [ 7.01473938,  1.28746666],
           [ 2.92383029,  1.28746666],
           [ 0.5       ,  2.95454545],
           [ 0.5       ,  7.04545455],
           [ 4.03665944,  7.59970419],
           [ 8.12756853,  7.59970419],
           [ 8.68181818,  9.5       ],
           [ 4.59090909,  9.5       ],
           [ 0.5       ,  9.5       ]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('h', 'integer', range_='[2;inf)')
        _numeric('w', 'integer', range_='[2;inf)')
        _numeric('scan_length', 'floating',
                 range_='[{};inf)'.format(_min_scan_length))
        _numeric('num_points', 'integer',
                 range_='[{};inf)'.format(_min_num_points))
        _numeric('discrete', 'integer', range_='[2;inf)', ignore_none=True)
        _numeric('seed', 'integer', range_='[0;inf)', ignore_none=True)

    validate_input()

    coords = random_line_sample_surface(float(h - 1), float(w - 1),
                                        scan_length, float(num_points - 1),
                                        1.0, discrete=discrete, seed=seed)
    coords = coords + 0.5

    return coords


def random_line_sample_surface(l, w, speed, sample_rate, time, discrete=None,
                               seed=None):
    """
    Sample a surface area using a set of random straight lines.

    The coordinates (in units of meters) resulting from sampling an image of
    size `l` times `w` using a pattern based on a set of random straight lines
    are determined.  The scanned path is determined from the probe `speed` and
    the scan `time`. If `discrete` is set, it specifies the finite number of
    equally spaced lines from which the scan lines are be chosen at random.
    For reproducible results, the `seed` may be used to specify a fixed seed of
    the random number generator.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.
    discrete : int or None, optional
        The number of equally spaced lines from which the scan lines are chosen
        (the default is None, which implies that no discritisation is used).
    seed : int or None, optional
        The seed used for the random number generator (the defaul is None,
        which implies that the random number generator is not seeded).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the length `l` is measured along the
    y-axis.

    Each of the scanned lines span the entire width of the image with the
    exception of the last line that may only be partially scanned if the
    `speed` and `time` implies this. The top and bottom lines of the image are
    always included in the scan and are not included in the `discrete` number
    of lines.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import random_line_sample_surface
    >>> l = 2e-6
    >>> w = 2e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> seed = 6021
    >>> np.set_printoptions(suppress=True)
    >>> random_line_sample_surface(l, w, speed, sample_rate, time, seed=seed)
    array([[ 0.        ,  0.        ],
           [ 0.00000067,  0.        ],
           [ 0.00000133,  0.        ],
           [ 0.000002  ,  0.        ],
           [ 0.000002  ,  0.00000067],
           [ 0.000002  ,  0.00000133],
           [ 0.00000158,  0.00000158],
           [ 0.00000091,  0.00000158],
           [ 0.00000024,  0.00000158],
           [ 0.        ,  0.000002  ],
           [ 0.00000067,  0.000002  ],
           [ 0.00000133,  0.000002  ],
           [ 0.000002  ,  0.000002  ]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('l', 'floating', range_='[{};inf)'.format(_min_l))
        _numeric('w', 'floating', range_='[{};inf)'.format(_min_w))
        _numeric('speed', 'floating', range_='[{};inf)'.format(_min_speed))
        _numeric('sample_rate', 'floating',
                 range_='[{};inf)'.format(_min_sample_rate))
        _numeric('time', 'floating', range_='[{};inf)'.format(_min_time))
        _numeric('discrete', 'integer', range_='[2;inf)', ignore_none=True)
        _numeric('seed', 'integer', range_='[0;inf)', ignore_none=True)

        if (speed * time - 2 * w - l) / w <= -1:
            # Estimated number of lines in addition to top and bottom lines
            # must exceed -1 to avoid drawing a negative number of lines at
            # random.
            msg = ('The value of >>(speed * time - 2 * w - l) / w<<, {!r}, '
                   'must be > -1.')
            raise ValueError(msg.format((speed * time - 2 * w - l) / w))

    validate_input()

    if seed is not None:
        np.random.seed(seed)

    num_lines = int(np.floor((speed * time - l) / w))

    if discrete is None:
        lines = np.sort(np.random.rand(num_lines - 2) * l)
    else:
        possible_lines = l / (discrete + 1) * np.arange(1, discrete + 1)

        try:
            lines = np.sort(np.random.choice(
                possible_lines, size=num_lines - 2, replace=False))
        except ValueError:
            raise ValueError('The number of Discrete lines must be large ' +
                             'enough to contain the entire scan path. With ' +
                             'the current settings, a minimun of '
                             '{!r} lines are required.'.format(num_lines - 2))

    coords = np.zeros((2 * num_lines, 2))
    coords[1::4, 0] = coords[2::4, 0] = w
    coords[2:-2:2, 1] = coords[3:-2:2, 1] = lines
    coords[-2:, 1] = l

    return _util.sample_lines(coords, speed, sample_rate, time)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing AFM image reconstruction and analysis of reconstructed images.

Routine listings
----------------
analyse(x, Phi, Psi)
    Sample an image, reconstruct it, and analyse the reconstructed image.
reconstruct(y, Phi, Psi)
    Reconstruct an image from compressively sensed measurements.

"""

from __future__ import division

from magni.afm import config as _conf
from magni.cs.reconstruction import amp as _amp
from magni.cs.reconstruction import gamp as _gamp
from magni.cs.reconstruction import iht as _iht
from magni.cs.reconstruction import it as _it
from magni.cs.reconstruction import sl0 as _sl0
from magni.imaging import evaluation as _eval
from magni.imaging import visualisation as _visualisation
from magni.utils.matrices import MatrixCollection as _MatrixC
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


def analyse(x, Phi, Psi, **kwargs):
    """
    Sample an image, reconstruct it, and analyse the reconstructed image.

    Parameters
    ----------
    x : numpy.ndarray
        The original image vector.
    Phi : magni.utils.matrices.Matrix or numpy.matrix
        The measurement matrix.
    Psi : magni.utils.matrices.Matrix or numpy.matrix
        The dictionary.

    Returns
    -------
    x : numpy.ndarray
        The reconstructed image vector.

    See Also
    --------
    magni.afm.config : Configuration options.
    magni.imaging.evaluation : Image reconstruction quality evaluation.

    Notes
    -----
    Additional \*\*kwargs are passed on to the reconstruction algorithm.

    Examples
    --------
    Prior to the actual example, data is loaded and a measurement matrix and a
    dictionary are defined. First, the example MI file provided with the
    package is loaded:

    >>> import os, numpy as np, magni
    >>> from magni.afm.reconstruction import analyse
    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> if os.path.isfile(path):
    ...     mi_file = magni.afm.io.read_mi_file(path)
    ...     mi_buffer = mi_file.get_buffer('Topography')[0]
    ...     mi_data = mi_buffer.data
    ...     x = magni.imaging.mat2vec(mi_data)

    Next, a measurement matrix is defined. This matrix is equal to the matrix
    generated by running `np.eye(len(x))[::2, :]` but for speed, the matrix is
    instead defined with fast operations:

    >>> def Phi_A(x):
    ...     y = x[::2]
    ...     return y
    >>> def Phi_T(y):
    ...     x = np.zeros((2 * len(y), 1))
    ...     x[::2] = y
    ...     return x
    >>> if os.path.isfile(path):
    ...     Phi = magni.utils.matrices.Matrix(Phi_A, Phi_T, (),
    ...                                       (int(len(x) / 2), len(x)))

    Next, a dictionary is defined. This dictionary is the DCT basis likewise
    defined with fast operations:

    >>> if os.path.isfile(path):
    ...     Psi = magni.imaging.dictionaries.get_DCT(mi_data.shape)

    Finally, the actual example:

    >>> if os.path.isfile(path):
    ...     print('MSE: {:.2f}, PSNR: {:.2f}'.format(*analyse(x, Phi, Psi)))
    ... else:
    ...     print('MSE: 0.24, PSNR: 6.22')
    MSE: 0.24, PSNR: 6.22

    """

    @_decorate_validation
    def validate_input():
        _numeric('x', ('integer', 'floating', 'complex'), shape=(-1, 1))
        _numeric('Phi', ('integer', 'floating', 'complex'),
                 shape=(-1, x.shape[0]))
        _numeric('Psi', ('integer', 'floating', 'complex'),
                 shape=(x.shape[0], -1))

    validate_input()

    y = Phi.dot(x)
    x_hat = reconstruct(y, Phi, Psi, **kwargs)

    x_scaled = _visualisation.stretch_image(x, 1.0)
    x_hat_scaled = _visualisation.stretch_image(x_hat, 1.0)

    mse = _eval.calculate_mse(x_scaled, x_hat_scaled)
    psnr = _eval.calculate_psnr(x_scaled, x_hat_scaled, 1.0)

    return (mse, psnr)


def reconstruct(y, Phi, Psi, **kwargs):
    """
    Reconstruct an image from compressively sensed measurements.

    Parameters
    ----------
    y : numpy.ndarray
        The measurement vector.
    Phi : magni.utils.matrices.Matrix or numpy.matrix
        The measurement matrix.
    Psi : magni.utils.matrices.Matrix or numpy.matrix
        The dictionary.

    Returns
    -------
    x : numpy.ndarray
        The reconstructed image vector.

    See Also
    --------
    magni.afm.config : Configuration options.
    magni.cs.reconstruction : Compressed sensing reconstruction algorithms.

    Notes
    -----
    Additional \*\*kwargs are passed on to the reconstruction algorithm.

    Examples
    --------
    Prior to the actual example, data is loaded and a measurement matrix and a
    dictionary are defined. First, the example MI file provided with the
    package is loaded:

    >>> import os, numpy as np, magni
    >>> from magni.afm.reconstruction import reconstruct
    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> if os.path.isfile(path):
    ...     mi_file = magni.afm.io.read_mi_file(path)
    ...     mi_buffer = mi_file.get_buffer('Topography')[0]
    ...     mi_data = mi_buffer.data
    ...     x = magni.imaging.mat2vec(mi_data)

    Next, a measurement matrix is defined. This matrix is equal to the matrix
    generated by running `np.eye(len(x))[::2, :]` but for speed, the matrix is
    instead defined with fast operations:

    >>> def Phi_A(x):
    ...     y = x[::2]
    ...     return y
    >>> def Phi_T(y):
    ...     x = np.zeros((2 * len(y), 1))
    ...     x[::2] = y
    ...     return x
    >>> if os.path.isfile(path):
    ...     Phi = magni.utils.matrices.Matrix(Phi_A, Phi_T, (),
    ...                                       (int(len(x) / 2), len(x)))

    Next, a dictionary is defined. This dictionary is the DCT basis likewise
    defined with fast operations:

    >>> if os.path.isfile(path):
    ...     Psi = magni.imaging.dictionaries.get_DCT(mi_data.shape)

    Finally, the actual example:

    >>> if os.path.isfile(path):
    ...     y = Phi.dot(x)
    ...     print('Maximum absolute pixel error: {:.3f}'
    ...           .format(np.abs(reconstruct(y, Phi, Psi) - x).max()))
    ... else:
    ...     print('Maximum absolute pixel error: 0.960')
    Maximum absolute pixel error: 0.960

    """

    @_decorate_validation
    def validate_input():
        _numeric('y', ('integer', 'floating', 'complex'), shape=(-1, 1))
        _numeric('Phi', ('integer', 'floating', 'complex'),
                 shape=(y.shape[0], -1))
        _numeric('Psi', ('integer', 'floating', 'complex'),
                 shape=(Phi.shape[1], -1))

        valid_kwargs = {'amp': [],
                        'gamp': ['A_asq'],
                        'iht': [],
                        'it': [],
                        'sl0': []}

        for kwarg in kwargs:
            if kwarg not in valid_kwargs[_conf['algorithm']]:
                raise TypeError(
                    'The {!r} algorithm does not '.format(_conf['algorithm']) +
                    'support the keyword argument >>{!r}<<'.format(kwarg))

    validate_input()

    A = _MatrixC((Phi, Psi))
    algorithm = _conf['algorithm']

    if algorithm == 'amp':
        algorithm = _amp.run
    elif algorithm == 'gamp':
        algorithm = _gamp.run
    elif algorithm == 'iht':
        algorithm = _iht.run
    elif algorithm == 'it':
        algorithm = _it.run
    elif algorithm == 'sl0':
        A = A.A
        algorithm = _sl0.run

    alpha = algorithm(y, A, **kwargs)
    x = Psi.dot(alpha)

    return x



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing custom data types.

Routine listings
----------------
ClassProperty(property)
    Class property.
ReadOnlyDict(collections.OrderedDict)
    Read-only ordered dict.

"""

from __future__ import division
from collections import OrderedDict as _OrderedDict


class ClassProperty(property):
    """
    Class property.

    The present class is a combination of the built-in property type and the
    built-in classmethod type. That is, it is a property which is invoked on a
    class rather than an instance but is available to both the class and its
    instances.

    Parameters
    ----------
    fget : function, optional
        The getter function of the property. (the default is None, which
        implies that the property cannot be read)
    fset : function, optional
        The setter function of the property. (the default is None, which
        implies that the property cannot be written)
    fdel : function, optional
        The deleter function of the property. (the default is None, which
        implies that the property cannot be deleted)
    doc : string, optional
        The docstring of the property. (the default is None, which implies that
        the docstring of the getter function, if any, is used)

    See Also
    --------
    property : Superclass from which all behaviour is inherited or extended.

    Examples
    --------
    The following example illustrates the difference between regular properties
    and class properties. First, the example class is defined and instantiated:

    >>> from magni.utils.types import ClassProperty
    >>> class Example(object):
    ...     _x_class = 0
    ...     def __init__(self, x):
    ...         self._x_instance = x
    ...     @ClassProperty
    ...     def x_class(class_):
    ...         return class_._x_class
    ...     @property
    ...     def x_instance(self):
    ...         return self._x_instance
    >>> example = Example(1)

    The regular read-only property works on the instance:

    >>> print('{!r}'.format(example.x_instance))
    1
    >>> print('Is property? {!r}'.format(isinstance(Example.x_instance,
    ...                                             property)))
    Is property? True

    The class property, on the other hand, works on the class:

    >>> print('{!r}'.format(example.x_class))
    0
    >>> print('Is property? {!r}'.format(isinstance(Example.x_class,
    ...                                             property)))
    Is property? False
    >>> print('{!r}'.format(Example.x_class))
    0

    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        property.__init__(self, fget, fset, fdel, doc)

    def __get__(self, obj, class_=None):
        """
        The get method of the property.

        Parameters
        ----------
        obj : object
            The instance on which the property is requested.
        class_ : type, optional
            The class on which the property is requested. (the default is None,
            which implies that the class is retrieved from `obj`)

        Returns
        -------
        value : None
            The value of the property.

        See Also
        --------
        property.__get__ : The method being extended.

        """

        if class_ is None:
            class_ = type(obj)

        return property.__get__(self, class_)

    def __set__(self, obj, value):
        """
        The set method of the property.

        Parameters
        ----------
        obj : object
            The instance on which the property is requested.
        value : None
            The value which should be assigned to the property.

        See Also
        --------
        property.__set__ : The method being extended.

        """

        return property.__set__(self, type(obj), value)

    def __delete__(self, obj):
        """
        The delete method of the property.

        Parameters
        ----------
        obj : object
            The instance on which the property is requested.

        See Also
        --------
        property.__delete__ : The method being extended.

        """

        return property.__delete__(self, type(obj))


class ReadOnlyDict(_OrderedDict):
    """
    Read-only ordered dict.

    The present ordered dict subclass has its non-read-only methods disabled.

    See Also
    --------
    collections.OrderedDict : Superclass from which all read-only behaviour is
        inherited.

    Examples
    --------
    This ordered dict subclass exposes the same interface as the OrderedDict
    class when not using methods that alter the dict.

    >>> from magni.utils.types import ReadOnlyDict
    >>> d = ReadOnlyDict(key='value')
    >>> for item in d.items():
    ...     print('{!r}'.format(item))
    ('key', 'value')

    However, any attempt to assign another value to the property raises an
    exception:

    >>> try:
    ...     d['key'] = 'new value'
    ... except Exception:
    ...     print('An exception occured.')
    An exception occured.

    """

    def __init__(self, *args, **kwargs):
        self._disabled = True
        _OrderedDict.__init__(self, *args, **kwargs)
        self._disabled = False

    def __delitem__(self, name):
        """
        Prevent deletion of items.

        Parameters
        ----------
        name : str
            The name of the item to be deleted.

        """

        if self._disabled:
            _OrderedDict.__delitem__(self, name)
        else:
            raise AttributeError('{!r} objects are read-only.'
                                 .format(self.__class__))

    def __getattribute__(self, name):
        """
        Return the requested attribute unless it is a non-read-only method.

        The purpose of this overwrite is to disable access to the following
        non-read-only dict methods: `clear`, `pop`, `popitem`, `setdefault`,
        and `update`. The first two methods are disabled otherwise.

        Parameters
        ----------
        name : str
            The name of the requested attribute.

        Returns
        -------
        attribute : None
            The requested attribute.

        Notes
        -----
        `__getattribute__` is implicitly called, when the attribute of an
        object is accessed as `object.attribute`.

        """

        if name in ('clear', 'pop', 'popitem', 'setdefault', 'update'):
            if self._disabled:
                value = _OrderedDict.__getattribute__(self, name)
            else:
                raise TypeError('{!r} objects are read-only.'
                                .format(self.__class__))
        else:
            value = _OrderedDict.__getattribute__(self, name)

        return value

    def __setitem__(self, name, value):
        """
        Prevent overwrite of items.

        Parameters
        ----------
        name : str
            The name of the item to be overwritten.
        value : None
            The value to be written.

        """

        if self._disabled:
            _OrderedDict.__setitem__(self, name, value)
        else:
            raise AttributeError('{!r} objects are read-only.'
                                 .format(self.__class__))



"""..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module for wrapping the Magni IPython Notebook examples.

**This module is based on the "ipnbdoctest.py" script by Benjamin
Ragan-Kelley (MinRK)**, source: https://gist.github.com/minrk/2620735.

This assumes comparison of IPython Notebooks in nbformat.v3

"""

from __future__ import division, print_function
import base64
from datetime import datetime
import os
import platform
import shutil
import subprocess
import sys
import unittest
import types
import warnings
try:
    from Queue import Empty  # Python 2
except ImportError:
    from queue import Empty  # Python 3
try:
    from StringIO import StringIO as BytesIO  # Python 2
except ImportError:
    from io import BytesIO  # Python 3

import numpy as np
from pkg_resources import parse_version
import scipy.misc

import magni

# The great "support IPython 2, 3, 4" strat begins
import IPython
try:
    import jupyter
except ImportError:
    jupyter_era = False
else:
    jupyter_era = True

if jupyter_era:
    # Jupyter / IPython 4.x
    from jupyter_client import KernelManager
    from nbformat import reads, NotebookNode

    def mod_reads(file_):
        return reads(file_, 3)  # Read notebooks as v3

else:
    from IPython.kernel import KernelManager
    with warnings.catch_warnings():
        warnings.simplefilter('error')
        try:
            # IPython 2.x
            from IPython.nbformat.current import reads, NotebookNode

            def mod_reads(file_):
                return reads(file_, 'json')

        except UserWarning:
            # IPython 3.x
            from IPython.nbformat import reads, NotebookNode

            def mod_reads(file_):
                return reads(file_, 3)  # Read notebooks as v3

# End of the great "support IPython 2, 3, 4" strat

# Test for freetype library version
try:
    if parse_version(
            subprocess.check_output(
                ['freetype-config', '--ftversion']).decode().strip()
            ) <= parse_version('2.5.2'):
        _skip_display_data_tests = False
    else:
        _skip_display_data_tests = True
except OSError:
    _skip_display_data_tests = True

if _skip_display_data_tests:
    warnings.warn('Skipping display data ipynb tests.', RuntimeWarning)


class _Meta(type):
    """
    Identification of IPython Notebook examples and construction of test class.

    """

    def __new__(class_, name, bases, attrs):
        path = magni.__path__[0].rsplit(os.sep, 1)[0]
        path = path + os.path.sep + 'examples' + os.path.sep

        for filename in os.listdir(path):
            if (platform.system() == 'Darwin' and
                    sys.version_info.major == 3 and
                    sys.version_info.minor == 3 and
                    filename == 'utils-multiprocessing.ipynb'):
                # Skip the broken multiprocessing on OSX Python 3.3 since case
                warnings.warn(
                    'Skipping multiprocessing ipynb test.', RuntimeWarning)
                continue
            if filename[-6:] == '.ipynb':
                name = 'test_' + filename[:-6].replace('-', '_')
                func = attrs['_run_example']
                func = types.FunctionType(func.__code__, func.__globals__,
                                          name, (path + filename,))
                func.__doc__ = func.__doc__.format(filename)
                attrs[name] = func

        return type.__new__(class_, name, bases, attrs)


# For python 2 and 3 compatibility
class _Hack(_Meta):
    def __new__(class_, name, bases, attrs):
        return _Meta(name, (unittest.TestCase,), attrs)


_TestCase = type.__new__(_Hack, 'temp', (), {})


@unittest.skipIf(
    parse_version(IPython.__version__) <= parse_version('3.0') and
    sys.version_info.major == 2 and
    platform.system() == 'Darwin', 'Due to a problem in IPython, the Magni ' +
    'Notebook example tests stall on Mac OSX with IPython 2 and Python 2.')
class TestIPythonExamples(_TestCase):
    """
    Test of Ipython Notebook examples for equality of output to reference.

    """

    def setUp(self):
        """
        Identify IPython Notebook examples to run.

        """

        path = magni.__path__[0].rsplit(os.sep, 1)[0]
        path = path + os.path.sep + 'examples' + os.path.sep
        files_to_copy = ['example.mi', 'data.hdf5', 'display.py']

        for cfile in files_to_copy:
            shutil.copy(os.path.join(path, cfile), '.')

    def _run_example(self, ipynb):
        """
        Test of {} Magni IPython Notebook example.

        """

        with open(ipynb) as f_ipynb:
            notebook = mod_reads(f_ipynb.read())

        notebook_result = _check_ipynb(notebook)
        passed, successes, failures, errors, report = notebook_result

        self.assertTrue(passed, msg=report)
        error_msg = ('Magni IPython Notebook example status:\n' +
                     'Successes: {}, Failures: {}, Errors: {}').format(
                         successes, failures, errors)
        self.assertEqual(errors + failures, 0, msg=error_msg)


def _check_ipynb(notebook):
    """
    Check an IPython Notebook for matching input and output.

    Each cell input in the `notebook` is executed and the result is compared
    to the cell output saved in the `notebook`.

    Parameters
    ----------
    notebook : IPython.nbformat.current.NotebookNode
        The notebook to check for matching input and output.

    Returns
    -------
    passed : Bool
        The indicator of a successful check (or not).
    sucessess : int
        The number of cell outputs that matched.
    failures : int
        The number of cell outputs that failed to match.
    errors : int
        The number of cell executions that resulted in errors.
    report : str
        The report detailing possible failures and errors.

    """

    kernel_manager = KernelManager()
    kernel_manager.start_kernel()
    kernel_client = kernel_manager.client()
    kernel_client.start_channels()

    try:
        # IPython 3.x
        kernel_client.wait_for_ready()
        iopub = kernel_client
        shell = kernel_client
    except AttributeError:
        # Ipython 2.x
        # Based on https://github.com/paulgb/runipy/pull/49/files
        iopub = kernel_client.iopub_channel
        shell = kernel_client.shell_channel
        shell.get_shell_msg = shell.get_msg
        iopub.get_iopub_msg = iopub.get_msg

    successes = 0
    failures = 0
    errors = 0

    report = ''
    for worksheet in notebook.worksheets:
        for cell in worksheet.cells:
            if cell.cell_type == 'code':
                try:
                    test_results = _execute_cell(cell, shell, iopub)
                except RuntimeError as e:
                    report += ('{!s} in cell number: {}'
                               .format(e, cell.prompt_number))
                    errors += 1
                    break

                identical_output = all(
                    [_compare_cell_output(test_result, reference)
                     for test_result, reference in
                     zip(test_results, cell.outputs)])

                if identical_output:
                    successes += 1
                else:
                    failures += 1

                    try:
                        str_test_results = [
                            '(for out {})\n'.format(k) + '\n'.join(
                                [' : '.join([str(key), str(val)])
                                 for key, val in t.items()
                                 if key not in ('metadata', 'png')]
                            ) for k, t in enumerate(test_results)]
                        str_cell_outputs = [
                            '(for out {})\n'.format(k) + '\n'.join(
                                [' : '.join([str(key), str(val)])
                                 for key, val in t.items()
                                 if key not in ('metadata', 'png')]
                            ) for k, t in enumerate(cell.outputs)]
                    except TypeError as e:
                        report += 'TypeError in ipynb_examples test\n\n'
                        for entry in cell.outputs:
                            if 'traceback' in entry.keys():
                                for item in entry['traceback']:
                                    report += str(item) + '\n'
                    else:
                        report += '\n' * 2 + '~' * 40
                        report += (
                            '\nFailure in {}:{}\nGot: {}\n\n\nExpected: {}'
                        ).format(notebook.metadata.name,
                                 cell.prompt_number,
                                 '\n'.join(str_test_results),
                                 '\n'.join(str_cell_outputs))

    kernel_client.stop_channels()
    kernel_manager.shutdown_kernel()

    passed = not (failures or errors)

    return passed, successes, failures, errors, report


def _compare_cell_output(test_result, reference):
    """
    Compare a cell test output to a reference output.

    Parameters
    ----------
    test_results : IPython.nbformat.current.NotebookNode
        The cell test result that must be compared to the reference.
    reference : IPython.nbformat.current.NotebookNode
        The reference cell output to compare to.

    Returns
    -------
    comparison_result : bool
        The indicator of equality between the test output and the reference.

    """

    skip_compare = ['traceback', 'latex', 'prompt_number']

    if _skip_display_data_tests:
        # Skip graphics comparison
        skip_compare.append('png')

    if test_result['output_type'] == 'display_data':
        # Prevent comparison of matplotlib figure instance memory addresses
        skip_compare.append('text')
        skip_compare.append('metadata')

    for key in reference:

        if key not in test_result:
            raise Exception(str(reference) + '!!!!!' + str(test_result))
            return False
        elif key not in skip_compare:
            if key == 'text':
                if test_result[key].strip() != reference[key].strip():
                    return False
            elif key == 'png':
                reference_img = reference[key]
                test_img = test_result[key]
                if not _compare_images(reference_img, test_img):
                    return False
            else:
                if test_result[key] != reference[key]:
                    return False

    return True


def _compare_images(reference_img, test_img):
    """
    Compare reference and test image to determine if they depict the same.

    Two images are considered to depict the same unless:

    - The image shapes differ
    - The number of differences in non-transparant pixel values which are not
      (likely to be) part of the image border exceeds 2.

    Parameters
    ----------
    reference_img : str
        The base64 encoded reference image.
    test_img : str
        The base64 encoded test image.

    Returns
    -------
    comparison_result : bool
        The idenfifier of a positive match, i.e. True if images are the same.

    """

    ref_png = base64.b64decode(reference_img)
    ref_ndarray = scipy.misc.imread(BytesIO(ref_png))
    cmp_png = base64.b64decode(test_img)
    cmp_ndarray = scipy.misc.imread(BytesIO(cmp_png))

    # check shape of images
    if cmp_ndarray.shape != ref_ndarray.shape:
        print('Image shapes differ')
        return False

    # mask of channels in pixels with different values
    diff = cmp_ndarray != ref_ndarray
    # mask of pixels with different values
    diff = np.any(diff, axis=2)
    # mask of non-transparent pixels with different values
    diff = diff * np.bool_(ref_ndarray[:, :, 3])

    # check if all non-transparent pixels match
    if diff.sum() == 0:
        # Accept difference in tranparent pixels
        return True

    # The rest is all about checking if (it is likely to be) only the image
    # border that has changed. The border may render differently across
    # matplotlib versions.

    # mask of black pixels
    mask = ((ref_ndarray[:, :, 0] == 0) *
            (ref_ndarray[:, :, 1] == 0) *
            (ref_ndarray[:, :, 2] == 0) *
            (ref_ndarray[:, :, 3] == 255))

    # lookup table of the top most connected black pixel of the
    # looked up pixel
    C_N = np.zeros(mask.shape, dtype=np.int16)

    for i in range(0, mask.shape[0] - 1):
        C_N[i + 1, :] = np.logical_not(mask[i, :]) * i + mask[i, :] * C_N[i, :]

    # lookup table of the right most connected black pixel of the
    # looked up pixel
    C_E = np.zeros(mask.shape, dtype=np.int16)

    for i in range(mask.shape[1] - 1, 0, -1):
        C_E[:, i - 1] = np.logical_not(mask[:, i]) * i + mask[:, i] * C_E[:, i]

    # lookup table of the bottom most connected black pixel of the
    # looked up pixel
    C_S = np.zeros(mask.shape, dtype=np.int16)

    for i in range(mask.shape[0] - 1, 0, -1):
        C_S[i - 1, :] = np.logical_not(mask[i, :]) * i + mask[i, :] * C_S[i, :]

    # lookup table of the left most connected black pixel of the
    # looked up pixel
    C_W = np.zeros(mask.shape, dtype=np.int16)

    for i in range(0, mask.shape[1] - 1):
        C_W[:, i + 1] = np.logical_not(mask[:, i]) * i + mask[:, i] * C_W[:, i]

    # coordinates of non-transparent pixels with different values
    points = np.nonzero(diff)
    points = np.int32(points + (np.zeros(points[0].shape),)).T

    # loop over non-transparent pixels with different values
    for i, point in enumerate(points):
        y, x = point[:2]

        # find other non-transparent pixels with different values
        # ... with the same y-coordinate
        matches_y = np.nonzero(points[:, 0] == y)[0]
        # ... with an x-coordinate at least 10 pixels away
        matches_y = matches_y[np.abs(points[matches_y, 1] - x) > 10]
        # ... which is connected by black pixels
        matches_y = matches_y[
            (points[matches_y, 1] >= C_W[y, x]) *
            (points[matches_y, 1] <= C_E[y, x])]

        # find other non-transparent pixels with different values
        # ... with the same x-coordinate
        matches_x = np.nonzero(points[:, 1] == x)[0]
        # ... with a y-coordinate at least 10 pixels away
        matches_x = matches_x[np.abs(points[matches_x, 0] - y) > 10]
        # ... which is connected by black pixels
        matches_x = matches_x[
            (points[matches_x, 0] >= C_N[y, x]) *
            (points[matches_x, 0] <= C_S[y, x])]

        if len(matches_y) + len(matches_x) == 0:
            # this pixel cannot be the corner of a box
            break

        for j in matches_y:
            for k in matches_x:
                # loop over combinations of possible boxes
                y_test = points[k, 0]
                x_test = points[j, 1]

                if not C_W[y_test, x] <= x_test <= C_E[y_test, x]:
                    # one horizontal line of the box isn't black
                    continue

                if not C_N[y, x_test] <= y_test <= C_S[y, x_test]:
                    # one vertical line of the box isn't black
                    continue

                # the box is a box and the corners are flagged
                points[i, 2] = points[j, 2] = points[k, 2] = 1

    if points.shape[0] - np.sum(points[:, 2]) > 2:
        print('The images differ by {} pixels'.format(
            points.shape[0] - np.sum(points[:, 2])))

        # Save images and their difference for visual inspection
        fail_txt = 'Notebook test fail '
        utcnow = datetime.utcnow
        scipy.misc.imsave(fail_txt + str(utcnow()) + 'r' + '.png', ref_ndarray)
        scipy.misc.imsave(fail_txt + str(utcnow()) + 't' + '.png', cmp_ndarray)
        img_diff = cmp_ndarray - ref_ndarray
        scipy.misc.imsave(fail_txt + str(utcnow()) + 'd' + '.png', img_diff)

        return False

    return True


def _execute_cell(cell, shell, iopub, timeout=300):
    """
    Execute an IPython Notebook Cell and return the cell output.

    Parameters
    ----------
    cell : IPython.nbformat.current.NotebookNode
        The IPython Notebook cell to execute.
    shell : IPython.kernel.blocking.channels.BlockingShellChannel
        The shell channel which the cell is submitted to for execution.
    iopub : IPython.kernel.blocking.channels.BlockingIOPubChannel
        The iopub channel used to retrieve the result of the execution.
    timeout : int
        The number of seconds to wait for the execution to finish before giving
        up.

    Returns
    -------
    cell_outputs : list
        The list of NotebookNodes holding the result of the execution.

    """

    # Execute input
    shell.execute(cell.input)
    exe_result = shell.get_shell_msg(timeout=timeout)
    if exe_result['content']['status'] == 'error':
        raise RuntimeError('Failed to execute cell due to error: {!r}'.format(
            str(exe_result['content']['evalue'])))

    cell_outputs = list()

    # Poll for iopub messages until no more messages are available
    while True:
        try:
            msg = iopub.get_iopub_msg(timeout=0.5)
        except Empty:
            break

        msg_type = msg['msg_type']
        if msg_type in ('status', 'pyin', 'execute_input', 'execute_result'):
            continue

        content = msg['content']
        node = NotebookNode(output_type=msg_type)

        if msg_type == 'stream':
            node.stream = content['name']
            if 'text' in content:
                # v4 notebook format
                node.text = content['text']
            else:
                # v3 notebook format
                node.text = content['data']

            bug_text = 'Using Anaconda Cloud api site https://api.anaconda.org'
            if bug_text in node.text:
                # Ignore conda (spam) messages/warnings
                continue
        elif msg_type in ('display_data', 'pyout'):
            node['metadata'] = content['metadata']
            for mime, data in content['data'].items():
                attr = mime.split('/')[-1].lower()
                attr = attr.replace('+xml', '').replace('plain', 'text')
                setattr(node, attr, data)
            if msg_type == 'pyout':
                node.prompt_number = content['execution_count']
        elif msg_type == 'pyerr':
            node.ename = content['ename']
            node.evalue = content['evalue']
            node.traceback = content['traceback']
        else:
            raise RuntimeError('Unhandled iopub message of type: {}'.format(
                msg_type))

        cell_outputs.append(node)

    return cell_outputs



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing data container classes for .mi files.

Routine listings
----------------
image
    Module providing data container classes for .mi image files.
spectroscopy
    Module providing data container classes for .mi spectroscopy files.
BaseClass(object)
    Base class of every `magni.afm.types` data class.
File(BaseClass)
    Base class of the `magni.afm.types` file classes.
FileCollection(BaseClass)
    Data class for collections of File instances with identical settings.

"""

# the BaseClass, File, and FileCollection classes need to be imported first to
# avoid recursive imports
from magni.afm.types._util import BaseClass
from magni.afm.types._util import File
from magni.afm.types._util import FileCollection

from magni.afm.types import image
from magni.afm.types import spectroscopy



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the multiprocessing subpackage.

See also
--------
magni.utils.config.Configger : The Configger class used

Notes
-----
This module instantiates the `Configger` class provided by
`magni.utils.config`. The configuration options are the following:

max_broken_pool_restarts : int or None
    The maximum number of attempts at restarting the process pool upon a
    BrokenPoolError (the default is 0, which implies that the process pool
    may not be restarted). If set to None, the process pool may restart
    indefinitely.
prefer_futures : bool
    The indicator of whether or not to prefer the concurrent.futures module
    over the multiprocessing module (the default is False, which implies that
    the multiprocessing module is used).
re_raise_exceptions : bool
    A flag indicating if exceptions should be re-raised (the default is False,
    which implies that the exception are not re-raised). It is useful to set
    this to True if the processing pool supports proper exception handling as
    e.g. when using "futures".
silence_exceptions : bool
    A flag indicating if exceptions should be silenced (the default is False,
    which implies that exceptions are raised).
workers : int
    The number of workers to use for multiprocessing (the default is 0, which
    implies no multiprocessing).

See the notes for the `magni.utils.multiprocessing._processing.process`
function for more details about the `prefer_futures` and
`max_broken_pool_restarts` configuration parameters.

"""

from __future__ import division

from magni.utils.config import Configger as _Configger
from magni.utils.validation import validate_numeric as _numeric


configger = _Configger(
    {'max_broken_pool_restarts': 0,
     'prefer_futures': False,
     're_raise_exceptions': False,
     'silence_exceptions': False,
     'workers': 0},
    {'max_broken_pool_restarts': _numeric(None, 'integer', range_='[0;inf)',
                                          ignore_none=True),
     'prefer_futures': _numeric(None, 'boolean'),
     're_raise_exceptions': _numeric(None, 'boolean'),
     'silence_exceptions': _numeric(None, 'boolean'),
     'workers': _numeric(None, 'integer', range_='[0;inf)')})



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the public functions of the magni.imaging subpackage.

"""

from __future__ import division

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


def double_mirror(img, fftstyle=False):
    """
    Mirror image in both the vertical and horisontal axes.

    The image is mirrored around its upper left corner first in the horizontal
    axis and then in the vertical axis such that an image of four times the
    size of the original is returned. If `fftstyle` is True, the image is
    constructed such it would represent a fftshifted version of the mirrored
    `img` such that entry (0, 0) is the DC component.

    Parameters
    ----------
    img : ndarray
        The image to mirror.
    fftstyle : bool
        The flag that indicates if the fftstyle mirrored image is returned.

    Returns
    -------
    mirrored_img : ndarray
        The mirrored image.

    Examples
    --------
    For example, mirror a very simple 2-by-3 pixel image.

    >>> import numpy as np
    >>> from magni.imaging._util import double_mirror
    >>> img = np.arange(6).reshape(2, 3)
    >>> img
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> double_mirror(img)
    array([[5, 4, 3, 3, 4, 5],
           [2, 1, 0, 0, 1, 2],
           [2, 1, 0, 0, 1, 2],
           [5, 4, 3, 3, 4, 5]])
    >>> double_mirror(img, fftstyle=True)
    array([[0, 0, 0, 0, 0, 0],
           [0, 5, 4, 3, 4, 5],
           [0, 2, 1, 0, 1, 2],
           [0, 5, 4, 3, 4, 5]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', ('boolean', 'integer', 'floating', 'complex'),
                 shape=(-1, -1))
        _numeric('fftstyle', 'boolean')

    validate_input()

    if fftstyle:
        mirrored_img = np.zeros((img.shape[0] * 2, img.shape[1] * 2),
                                dtype=img.dtype)
        tmp = np.pad(img, ((img.shape[0] - 1, 0), (img.shape[1] - 1, 0)),
                     mode='reflect')
        mirrored_img[1:, 1:] = tmp
    else:
        mirrored_img = np.pad(img, ((img.shape[0], 0), (img.shape[1], 0)),
                              mode='symmetric')

    return mirrored_img


def get_inscribed_masks(img, as_vec=False):
    """
    Return a set of inscribed masks covering the image.

    Two masks are returned. One is the disc with radius equal to that of the
    inscribed circle for `img`. The other is the inscribed square of the first
    mask. If `as_vec` is True, the `img` must be a vector representation of the
    (matrix) image. In this case, the masks are also returned in vector
    representation.

    Parameters
    ----------
    img : ndarray
        The square image of even height/width which the masks should cover.
    as_vec : bool
        The indicator of whether or not to treat `img` as a vector instead of
        an image (the default is False, which implies that `img` is treated as
        a matrix.

    Returns
    -------
    cicle_mask : ndarray
        The inscribed cicle mask.
    square_mask : ndarray
        The inscribed square mask.

    Examples
    --------
    For example, get the inscribed masks of an 8-by-8 image:

    >>> import numpy as np
    >>> from magni.imaging._util import get_inscribed_masks, mat2vec
    >>> img = np.arange(64).reshape(8, 8)
    >>> circle_mask, square_mask = get_inscribed_masks(img)
    >>> np_printoptions = np.get_printoptions()
    >>> np.set_printoptions(formatter={'bool': lambda x: str(int(x))})
    >>> circle_mask
    array([[0, 1, 1, 1, 1, 1, 1, 0],
           [1, 1, 1, 1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 1],
           [0, 1, 1, 1, 1, 1, 1, 0]], dtype=bool)

    >>> square_mask
    array([[0, 0, 0, 0, 0, 0, 0, 0],
           [0, 1, 1, 1, 1, 1, 1, 0],
           [0, 1, 1, 1, 1, 1, 1, 0],
           [0, 1, 1, 1, 1, 1, 1, 0],
           [0, 1, 1, 1, 1, 1, 1, 0],
           [0, 1, 1, 1, 1, 1, 1, 0],
           [0, 1, 1, 1, 1, 1, 1, 0],
           [0, 0, 0, 0, 0, 0, 0, 0]], dtype=bool)

    Or get the same masks based on a vector:

    >>> img_vec = mat2vec(img)
    >>> c_vec_mask, s_vec_mask = get_inscribed_masks(img_vec, as_vec=True)
    >>> c_vec_mask
    array([[0],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [0],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [0],
           [1],
           [1],
           [1],
           [1],
           [1],
           [1],
           [0]], dtype=bool)
    >>> np.set_printoptions(**np_printoptions)

    """

    @_decorate_validation
    def validate_input():
        _numeric('as_vec', 'boolean')

        if not as_vec:
            _numeric('img', ('boolean', 'integer', 'floating', 'complex'),
                     shape=(-1, -1))
            if img.shape[0] != img.shape[1]:
                raise ValueError('The input image must be square')

            if np.mod(img.shape[0], 2) != 0:
                raise ValueError('The image height/width must be an even' +
                                 ' number of pixels')
        else:
            _numeric('img', ('boolean', 'integer', 'floating', 'complex'),
                     shape=(-1, 1))
            if not np.allclose(np.mod(np.sqrt(img.size), 1), 0):
                raise ValueError('The input image must be square')

            if np.mod(np.sqrt(img.shape[0]), 2) != 0:
                raise ValueError('The image height/width must be an even' +
                                 ' number of pixels')

    validate_input()

    if as_vec:
        r = np.sqrt(img.shape[0]) / 2
    else:
        r = img.shape[0] / 2

    x, y = np.meshgrid(*map(np.arange, (r, r)))

    circle_mask = double_mirror(x**2 + y**2 <= r**2)
    square_mask = double_mirror(np.maximum(x, y) <= r / np.sqrt(2))

    if as_vec:
        return mat2vec(circle_mask), mat2vec(square_mask)
    else:
        return circle_mask, square_mask


def mat2vec(x):
    """
    Reshape `x` from matrix to vector by stacking columns.

    Parameters
    ----------
    x : ndarray
        Matrix that should be reshaped to vector.

    Returns
    -------
    ndarray
        Column vector formed by stacking the columns of the matrix `x`.

    See Also
    --------
    vec2mat : The inverse operation

    Notes
    -----
    The returned column vector is C contiguous.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging._util import mat2vec
    >>> x = np.arange(4).reshape(2, 2)
    >>> x
    array([[0, 1],
           [2, 3]])
    >>> mat2vec(x)
    array([[0],
           [2],
           [1],
           [3]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('x', ('boolean', 'integer', 'floating', 'complex'),
                 shape=(-1, -1))

    validate_input()

    return x.T.reshape(-1, 1)


def vec2mat(x, mn_tuple):
    """
    Reshape `x` from column vector to matrix.

    Parameters
    ----------
    x : ndarray
        Matrix that should be reshaped to vector.
    mn_tuple : tuple
        A tuple (m, n) containing the parameters m, n as listed below.
    m : int
        Number of rows in the resulting matrix.
    n : int
        Number of columns in the resulting matrix.

    Returns
    -------
    ndarray
       Matrix formed by taking `n` columns of lenght `m` from the column vector
       `x`.

    See Also
    --------
    mat2vec : The inverse operation

    Notes
    -----
    The returned matrix is C contiguous.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging._util import vec2mat
    >>> x = np.arange(4).reshape(4, 1)
    >>> x
    array([[0],
           [1],
           [2],
           [3]])
    >>> vec2mat(x, (2, 2))
    array([[0, 2],
           [1, 3]])

    """

    m, n = mn_tuple

    @_decorate_validation
    def validate_input():
        _numeric('m', 'integer', range_='[1;inf)')
        _numeric('n', 'integer', range_='[1;inf)')
        _numeric('x', ('boolean', 'integer', 'floating', 'complex'),
                 shape=(m * n, 1))

    validate_input()

    return x.reshape(n, m).T



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions that may be used to annotate data.

Routine Listings
----------------
get_conda_info()
    Function that returns information about an Anaconda install.
get_datetime()
    Function that returns information about the current date and time.
get_git_revision(git_root_dir=None)
    Function that returns information about the current git revision.
get_file_hashes(path, blocksize=2**30)
    Function that returns the md5 and sha256 checksums of a file.
get_magni_config()
    Function that returns information about the current configuration of Magni.
get_magni_info()
    Function that returns genral information about Magni.
get_platform_info()
    Function that returns information about the platform used to run the code.

Notes
-----
The returned annotations are any nested level of dicts of dicts of strings.

"""

from __future__ import division
import contextlib
import datetime
import hashlib
import json
import os
import pkgutil
import platform
import pydoc
import re
import subprocess
import sys

from pkg_resources import parse_version as _parse_version

import magni as _magni
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric

__all__ = ['get_conda_info', 'get_datetime', 'get_git_revision',
           'get_file_hashes', 'get_magni_config', 'get_magni_info',
           'get_platform_info']


def get_conda_info():
    """
    Return a dictionary contianing information from Conda.

    `Conda <http://conda.pydata.org/>`_ is the package manager for the Anaconda
    scientific Python distribution. This function will return various
    information about the Anaconda installation on the system by querying the
    Conda package database.

    .. warning::

        THIS IS HIGHLY EXPERIMENTAL AND MAY BREAK WITHOUT FURTHER NOTICE.

    .. note::

        This only works with the conda root environment.

    Returns
    -------
    conda_info : dict
        Various information from conda (see notes below for further details).

    Notes
    -----
    If the Python intepreter is unable to locate and import the conda package,
    an empty dicionary is returned.

    The returned dictionary contains the same infomation that is returned by
    "conda info" in addition to an overview of the linked modules in the
    Anaconda installation as well as information about the "conda env" if it is
    available. Specifically, the returned dictionary has the following keys:

    * platform
    * conda_version
    * root_prefix
    * default_prefix
    * envs_dirs
    * package_cache
    * channels
    * config_file
    * linked_modules
    * env_export

    For conda < 4.2.0, it also includes the key:

    * is_foreign_system

    For conda >= 4.2.0, it also includes the keys:

    * python_version
    * conda_is_private
    * offline_mode

    Additionally, the returned dictionary has a key named *status*, which can
    have either of the following values:

    * 'Succeeded' (Everything seems to be OK)
    * 'Failed' (Something went wrong - a few details are incluede in the key)

    If "conda-env" is installed on the system, the `env_export` essentially
    holds the infomation from "conda env export -n root" as a dictionary. The
    information provided by this key partially overlaps with the infomation in
    the `linked_modules` and `modules_info` keys.

    """

    try:
        import conda
        import conda.config
        import conda.install
    except ImportError:
        return {'status': 'Failed: Conda not importable'}

    # Conda info + linked modules
    if _parse_version(conda.__version__) >= _parse_version('4.2.0'):
        try:
            import conda.models.channel
            import conda.base.context
        except ImportError:
            return {'status': 'Failed: Conda context not importable'}
        conda_channel_urls = conda.models.channel.prioritize_channels(
            conda.base.context.context.channels)

    else:
        # Ugly hack to silence the
        # "Using Anaconda Cloud api site https://api.anaconda.org"
        # message being sent to stderr by the binstar/anaconda client.
        with _silence_stderr():
            conda_channel_urls = conda.config.get_channel_urls()

    conda_info = {'platform': conda.config.subdir,
                  'conda_version': conda.__version__,
                  'root_prefix': conda.config.root_dir,
                  'default_prefix': conda.config.default_prefix,
                  'envs_dirs': json.dumps(conda.config.envs_dirs),
                  'package_cache': json.dumps(conda.config.pkgs_dirs),
                  'channels': json.dumps(conda_channel_urls),
                  'config_file': json.dumps(conda.config.rc_path)}

    # Ugly hack to silence the
    # "Using Anaconda Cloud api site https://api.anaconda.org"
    # message being sent to stderr by the binstar/anaconda client.
    with _silence_stderr():
        linked_modules = sorted(conda.install.linked(conda.config.root_dir))

    if _parse_version(conda.__version__) >= _parse_version('4.2.0'):
        conda_info['python_version'] = '.'.join(map(str, sys.version_info))
        conda_info['conda_is_private'] = json.dumps(
            conda.base.context.context.conda_private)
        conda_info['offline_mode'] = json.dumps(
            conda.base.context.context.offline)
    else:
        conda_info['is_foreign_system'] = json.dumps(
            bool(conda.config.foreign))

    modules_info = {module:
                    conda.install.is_linked(conda_info['root_prefix'], module)
                    for module in linked_modules}
    conda_info['modules_info'] = json.dumps(modules_info)
    conda_info['linked_modules'] = json.dumps(linked_modules)
    conda_info['status'] = 'Succeeded'

    # Conda env export
    try:
        import conda_env.env
        conda_info['env_export'] = json.dumps(
            conda_env.env.from_environment(
                'root', conda.config.root_dir).to_dict())
    except ImportError:
        conda_info['env_export'] = 'Failed: conda-env not available'

    return conda_info


def get_datetime():
    """
    Return a dictionary holding the current date and time.

    Returns
    -------
    date_time : dict
        The dictionary holding the current date and time.

    Notes
    -----
    The returned dictionary has the following keys:

    * today (date and time including timezone offset)
    * utcnow (UTC date and time)
    * pretty_utc (UTC date and time formatted according to current locale)
    * status

    The status entry informs about the success of the pretty_utc formatting.
    It has one of the follwing values:

    * Succeeded (Everything seems OK)
    * Failed (It was not possible to format the time)

    """

    date_time = {'today': repr(datetime.datetime.today()),
                 'utcnow': datetime.datetime.utcnow(),
                 'pretty_utc': '',
                 'status': 'Succeeded'}

    try:
        date_time['pretty_utc'] = datetime.datetime.strftime(
            date_time['utcnow'], '%c')

    except ValueError:
        date_time['status'] = 'Failed'

    date_time['utcnow'] = repr(date_time['utcnow'])

    return date_time


def get_git_revision(git_root_dir=None):
    """
    Return a dictionary containing information about the current git revision.

    Parameters
    ----------
    git_root_dir : str
        The path to the git root directory to get git revision for (the default
        is None, which implies that the git revision of the `magni` directory
        is returned).

    Returns
    -------
    git_revision : dict
        Information about the current git revision.

    Notes
    -----
    If the git revision extract succeeded, the returned dictionary has the
    following keys:

    * status (with value 'Succeeded')
    * tag (output of "git describe")
    * branch (output of "git describe --all")
    * remote (output of "git remote -v")

    If the git revision extract failed, the returned dictionary has the
    following keys:

    * status (with value 'Failed')
    * returncode (returncode from failing git command)
    * output (output from failing git command)

    The "git" commands are run in the git root directory.

    """

    @_decorate_validation
    def validate_input():
        _generic('git_root_dir', 'string', ignore_none=True)

    validate_input()

    cur_dir = os.getcwd()
    if git_root_dir is not None:
        try:
            os.chdir(git_root_dir)
        except (IOError, OSError):
            raise OSError(
                'The git_root_dir directory "{!r}".format(git_root_dir)' +
                'does not exist')
    else:
        os.chdir(os.path.split(_magni.__path__[0])[0])

    try:
        git_revision = {
            'tag': str(subprocess.check_output(
                ['git', 'describe'],
                stderr=subprocess.STDOUT)[:-1].decode()),
            'branch': str(subprocess.check_output(
                ['git', 'describe', '--all'],
                stderr=subprocess.STDOUT)[:-1].decode()),
            'remote': str(subprocess.check_output(
                ['git', 'remote', '-v'],
                stderr=subprocess.STDOUT)[:-1].decode()),
            'status': 'Succeeded'}

    except subprocess.CalledProcessError as e:
        try:
            e_output = e.output.decode()
        except AttributeError:
            e_output = e.output

        git_revision = {'status': 'Failed: CallProcessError',
                        'returncode': e.returncode,
                        'output': e_output}

    except OSError as e:
        try:
            e_strerror = e.strerror.decode()
        except AttributeError:
            e_strerror = e.strerror

        git_revision = {'status': 'Failed: OSError',
                        'errno': e.errno,
                        'strrror': e_strerror}

    os.chdir(cur_dir)

    return git_revision


def get_file_hashes(path, blocksize=2**30):
    """
    Return a dictionary with md5 and sha256 checksums of a file.

    Parameters
    ----------
    path : str
        The path to the file to checksum.
    blocksize : int
        The chunksize (in bytes) to read from the file one at a time.

    Returns
    -------
    file_hashes : dict
        The dictionary holding the md5 and sha256 hexdigests of the file.

    """

    @_decorate_validation
    def validate_input():
        _generic('path', 'string')
        _numeric('blocksize', 'integer', range_='[1;inf)')

    validate_input()

    md5sum = hashlib.md5()
    sha256sum = hashlib.sha256()

    with open(path, mode='rb') as f_handle:
        buf = f_handle.read(blocksize)
        while buf != ''.encode():
            md5sum.update(buf)
            sha256sum.update(buf)
            buf = f_handle.read(blocksize)

    file_hashes = {'md5sum': md5sum.hexdigest(),
                   'sha256sum': sha256sum.hexdigest()}

    return file_hashes


def get_magni_config():
    """
    Return a dictionary holding the current configuration of Magni.

    Returns
    -------
    magni_config : dict
        The dictionary holding the current configuration of Magni.

    Notes
    -----
    The returned dictionary has a key for each of the `config` modules in Magni
    and its subpackages. The value of a given key is a dictionary with the
    current configuration of the corresponding `config` module. Furthermore,
    the returned dictionary has a status key, which can have either of the
    following values:

    * Succeeded (The entire configuration was extracted)
    * Failed (It was not possible to get information from one or more modules)

    """

    packages = pkgutil.walk_packages(path=_magni.__path__,
                                     prefix=_magni.__name__ + '.')

    magni_config = dict()
    try:
        for importer, modname, ispkg in packages:
            if modname[-8:] == '._config':
                try:
                    settings = dict(eval('_' + modname).configger.items())
                except AttributeError:
                    # Skip base Configgers, e.g. cs.reconcstruction.config
                    pass
                else:
                    for setting in settings:
                        if not isinstance(settings[setting], str):
                            settings[setting] = repr(settings[setting])

                    magni_config[modname[:-7] + modname[-6:]] = settings

        magni_config['status'] = 'Succeeded'

    except AttributeError:
        magni_config['status'] = 'Failed'

    return magni_config


def get_magni_info():
    """
    Return a string representation of the output of help(magni).

    Returns
    -------
    magni_info : dict
        Information about magni.

    Notes
    -----
    The returned dictionary has a single key:

    * help_magni (a string representation of help(magni))

    """

    magni_info = pydoc.render_doc(_magni)
    magni_info, subs = re.subn(r'\x08([A-Z]|[a-z]|_)?', '', magni_info)

    return {'help_magni': magni_info}


def get_platform_info():
    """
    Return a dictionary containing information about the system platform.

    Returns
    -------
    platform_info : dict
        Various information about the system platform.

    See Also
    --------
    platform : The Python module used to query information about the system.

    Notes
    -----
    The returned dictionary has the following keys:

    * system
    * node
    * release
    * version
    * processor
    * python
    * libc
    * linux
    * mac_os
    * win32
    * status

    The linux/mac_os/win32 entries are "empty" if they are not applicable.

    If the processor information returned by `platform` is "empty", a query of
    `lscpu` is attempted in order to provide the necessary information.

    The status entry informs about the success of the queries. It has one of
    the follwing values:

    * 'All OK' (everything seems to be OK)
    * 'Used lscpu in processor query' (`lscpu` was used)
    * 'Processor query failed' (failed to get processor information)

    """

    platform_info = {'system': json.dumps(platform.system()),
                     'node': json.dumps(platform.node()),
                     'release': json.dumps(platform.release()),
                     'version': json.dumps(platform.version()),
                     'machine': json.dumps(platform.machine()),
                     'processor': json.dumps(platform.processor()),
                     'python': json.dumps(platform.python_version()),
                     'libc': json.dumps(platform.libc_ver()),
                     'linux': json.dumps(platform.linux_distribution()),
                     'mac_os': json.dumps(platform.mac_ver()),
                     'win32': json.dumps(platform.win32_ver()),
                     'status': 'All OK'}

    if platform_info['processor'] == '':
        try:
            platform_info['processor'] = str(
                subprocess.check_output(['lscpu']).decode())
            platform_info['status'] = 'Used lscpu in processor query'

        except (subprocess.CalledProcessError, OSError):
            platform_info['status'] = 'Processor query failed'

    return platform_info


@contextlib.contextmanager
def _silence_stderr():
    """
    A context manager for silencing stderr.

    """

    _stderr = sys.stderr
    with open(os.devnull, 'wb') as null:
        sys.stderr = null

    yield

    sys.stderr = _stderr



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests for `magni.imaging.evaluation`.

**Testing Strategy**

Three cases are considered for the calculation of the evaluation metrics:

* An identical array
* A constant offset to all elements of an array
* An array subject to additive white gaussian noise

For all three cases flat, 1D, and 2D arrays are tested.

"""

from __future__ import division
import unittest

import numpy as np

from magni.imaging.evaluation import (calculate_mse, calculate_psnr,
                                      calculate_retained_energy)


class TestMetrics(unittest.TestCase):
    """
    Test of the evaluation metrics.

    """

    def setUp(self):
        n = 10
        N = n ** 2

        np.random.seed(seed=6021)

        self.flat_array = np.random.permutation(1 / N * np.arange(N))
        self.d1_array = self.flat_array.copy().reshape(N, 1)
        self.d2_array = self.flat_array.copy().reshape(n, n)

        c = 0.27
        self.flat_const_offset = c * np.ones_like(self.flat_array)
        self.d1_const_offset = c * np.ones_like(self.d1_array)
        self.d2_const_offset = c * np.ones_like(self.d2_array)

        self.flat_noise = np.random.normal(scale=1/n, size=(N,))
        self.d1_noise = self.flat_noise.copy().reshape(N, 1)
        self.d2_noise = self.flat_noise.copy().reshape(n, n)

    def test_MSE(self):
        k = len(self.flat_array)
        x_zero = np.zeros(k)
        x_flat = np.arange(k)
        x_1d = np.arange(k).reshape(*self.d1_array.shape)
        x_2d = np.arange(k).reshape(*self.d2_array.shape)

        mse_const = np.mean(self.flat_const_offset ** 2)
        mse_noise = np.mean(self.flat_noise ** 2)

        # Integrity check of test setup
        self.assertGreater(mse_const, 1e-3)
        self.assertGreater(mse_noise, 1e-3)
        self.assertNotAlmostEqual(mse_const, mse_noise)

        # Identical arrays
        self.assertEqual(calculate_mse(x_zero, x_zero), 0)
        self.assertEqual(calculate_mse(x_2d, x_2d), 0)

        # Additve offset
        self.assertAlmostEqual(
            calculate_mse(x_flat, x_flat + self.flat_const_offset), mse_const)
        self.assertAlmostEqual(
            calculate_mse(x_1d, x_1d - self.d1_const_offset), mse_const)
        self.assertAlmostEqual(
            calculate_mse(x_2d, x_2d + self.d2_const_offset), mse_const)

        # Additive noise
        self.assertAlmostEqual(
            calculate_mse(x_flat, x_flat - self.flat_noise), mse_noise)
        self.assertAlmostEqual(
            calculate_mse(x_1d, x_1d + self.d1_noise), mse_noise)
        self.assertAlmostEqual(
            calculate_mse(x_2d, x_2d - self.d2_noise), mse_noise)

        # Fails
        self.assertNotAlmostEqual(
            calculate_mse(x_flat, x_flat + self.flat_noise), mse_const)
        self.assertNotAlmostEqual(
            calculate_mse(x_flat, x_flat - self.flat_const_offset), mse_noise)

    def test_PSNR(self):
        k = len(self.flat_array)
        x_zero = np.zeros(k)
        x_flat = np.arange(k)
        x_1d = np.arange(k).reshape(*self.d1_array.shape)
        x_2d = np.arange(k).reshape(*self.d2_array.shape)

        p = 1.42
        psnr_const = 10 * np.log10(p**2 / np.mean(self.flat_const_offset ** 2))
        psnr_noise = 10 * np.log10(p**2 / np.mean(self.flat_noise ** 2))

        # Integrity check of test setup
        self.assertGreater(psnr_const, 1e-3)
        self.assertGreater(psnr_noise, 1e-3)
        self.assertNotAlmostEqual(psnr_const, psnr_noise)

        # Identical arrays
        self.assertEqual(calculate_psnr(x_zero, x_zero, p), np.inf)
        self.assertEqual(calculate_psnr(x_2d, x_2d, p), np.inf)

        # Additve offset
        self.assertAlmostEqual(
            calculate_psnr(x_flat, x_flat + self.flat_const_offset, p),
            psnr_const)
        self.assertAlmostEqual(
            calculate_psnr(x_1d, x_1d - self.d1_const_offset, p), psnr_const)
        self.assertAlmostEqual(
            calculate_psnr(x_2d, x_2d + self.d2_const_offset, p), psnr_const)

        # Additive noise
        self.assertAlmostEqual(
            calculate_psnr(x_flat, x_flat - self.flat_noise, p), psnr_noise)
        self.assertAlmostEqual(
            calculate_psnr(x_1d, x_1d + self.d1_noise, p), psnr_noise)
        self.assertAlmostEqual(
            calculate_psnr(x_2d, x_2d - self.d2_noise, p), psnr_noise)

        # Fails
        self.assertNotAlmostEqual(
            calculate_psnr(x_flat, x_flat + self.flat_noise, p), psnr_const)
        self.assertNotAlmostEqual(
            calculate_psnr(x_flat, x_flat - self.flat_const_offset, p),
            psnr_noise)
        self.assertNotAlmostEqual(
            calculate_psnr(x_flat, x_flat + self.flat_noise, 25), psnr_noise)
        self.assertNotAlmostEqual(
            calculate_psnr(x_flat, x_flat + self.flat_noise, p*1.01),
            psnr_noise)

    def test_retained_energy(self):
        k = len(self.flat_array)
        x_zero = np.zeros(k)
        x_flat = np.ones(k)
        x_1d = np.ones(k).reshape(*self.d1_array.shape)
        x_2d = np.ones(k).reshape(*self.d2_array.shape)

        energy_const = 1/k * np.sum(self.flat_const_offset ** 2) * 100
        energy_noise = 1/k * np.sum(self.flat_noise ** 2) * 100

        # Integrity check of test setup
        self.assertGreater(energy_const, 1e-3)
        self.assertGreater(energy_noise, 1e-3)
        self.assertNotAlmostEqual(energy_const, energy_noise)

        # Identical arrays and zero case
        with self.assertRaises(ValueError):
            calculate_retained_energy(x_zero, x_zero)

        self.assertEqual(calculate_retained_energy(x_flat, x_flat), 100)
        self.assertEqual(calculate_retained_energy(x_flat, x_zero), 0)

        # Additve offset
        self.assertAlmostEqual(
            calculate_retained_energy(
                x_flat, self.flat_const_offset), energy_const)
        self.assertAlmostEqual(
            calculate_retained_energy(
                x_1d, self.d1_const_offset), energy_const)
        self.assertAlmostEqual(
            calculate_retained_energy(
                x_2d, self.d2_const_offset), energy_const)

        # Additive noise
        self.assertAlmostEqual(
            calculate_retained_energy(x_flat, self.flat_noise), energy_noise)
        self.assertAlmostEqual(
            calculate_retained_energy(x_1d, self.d1_noise), energy_noise)
        self.assertAlmostEqual(
            calculate_retained_energy(x_2d, self.d2_noise), energy_noise)

        # Fails
        self.assertNotAlmostEqual(
            calculate_retained_energy(x_flat, self.flat_noise), energy_const)
        self.assertNotAlmostEqual(
            calculate_mse(x_flat, self.flat_const_offset), energy_noise)
        with self.assertRaises(ValueError):
            calculate_retained_energy(x_zero, x_flat)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
spiral_sample_image(h, w, scan_length, num_points, rect_area=False)
    Function for spiral sampling an image.
spiral_sample_surface(l, w, speed, sample_rate, time, rect_area=False)
    Function for spiral sampling a surface.

"""

from __future__ import division

import numpy as np

from magni.imaging.measurements import _util
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['spiral_sample_image', 'spiral_sample_surface']

_min_l = _util.min_l
_min_w = _util.min_w
_min_speed = _util.min_speed
_min_sample_rate = _util.min_sample_rate
_min_time = _util.min_time
_min_scan_length = _util.min_scan_length
_min_num_points = _util.min_num_points


def spiral_sample_image(h, w, scan_length, num_points, rect_area=False):
    """
    Sample an image using an archimedean spiral pattern.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using an archimedean spiral pattern are determined. The
    `scan_length` determines the length of the path scanned whereas
    `num_points` indicates the number of samples taken on that path.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.
    rect_area : bool
        A flag indicating whether or not the full rectangular area is sampled
        (the default value is False which implies that the "corners" of the
        rectangular area are not sampled).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis. The width must equal the height for an archimedian spiral to make
    sense.

    If the `rect_area` flag is True, then it is assumed that the sampling
    continues outside of the rectangular area specified by `h` and `w` such
    that the "corners" of the rectangular area are also sampled. The sample
    points outside of the rectangular area are discarded and, hence, not
    returned.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import spiral_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> np.set_printoptions(suppress=True)
    >>> spiral_sample_image(h, w, scan_length, num_points)
    array([[ 6.28776846,  5.17074073],
           [ 3.13304898,  5.24133767],
           [ 6.07293751,  2.93873701],
           [ 6.99638041,  6.80851189],
           [ 2.89868434,  7.16724999],
           [ 2.35773914,  3.00320067],
           [ 6.41495385,  1.71018152],
           [ 8.82168896,  5.27557847],
           [ 6.34932919,  8.83624957],
           [ 2.04885699,  8.11199373],
           [ 0.6196052 ,  3.96939755]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('h', 'integer', range_='[2;inf)')
        _numeric('w', 'integer', range_='[2;inf)')

        if h != w:
            msg = ('The value of >>h<<, {!r}, must equal the value of >>w<<, '
                   '{!r}, for an archimedian spiral to make sense.')
            raise ValueError(msg.format(h, w))

        _numeric('scan_length', 'floating',
                 range_='[{};inf)'.format(_min_scan_length))
        _numeric('num_points', 'integer',
                 range_='[{};inf)'.format(_min_num_points))
        _numeric('rect_area', 'boolean')

    validate_input()

    coords = spiral_sample_surface(float(h - 1), float(w - 1),
                                   scan_length, float(num_points), 1.0,
                                   rect_area)
    coords = coords + 0.5

    return coords


def spiral_sample_surface(l, w, speed, sample_rate, time, rect_area=False):
    """
    Sample a surface area using an archimedean spiral pattern.

    The coordinates (in units of meters) resulting from sampling an area of
    size `l` times `w` using an archimedean spiral pattern are determined. The
    scanned path is determined from the probe `speed` and the scan `time`.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.
    rect_area : bool
        A flag indicating whether or not the full rectangular area is sampled
        (the default value is False which implies that the "corners" of the
        rectangular area are not sampled).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the length `l` is measured along the
    y-axis. The width must equal the length for an archimedian sprial to make
    sense.

    If the `rect_area` flag is True, then it is assumed that the sampling
    continues outside of the rectangular area specified by `l` and `w` such
    that the "corners" of the rectangular area are also sampled. The sample
    points outside of the rectangular area are discarded and, hence, not
    returned.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import spiral_sample_surface
    >>> l = 1e-6
    >>> w = 1e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> np.set_printoptions(suppress=True)
    >>> spiral_sample_surface(l, w, speed, sample_rate, time)
    array([[ 0.00000036,  0.00000046],
           [ 0.00000052,  0.00000071],
           [ 0.00000052,  0.00000024],
           [ 0.00000059,  0.00000079],
           [ 0.00000021,  0.00000033],
           [ 0.00000084,  0.00000036],
           [ 0.00000049,  0.0000009 ],
           [ 0.0000001 ,  0.00000036],
           [ 0.00000072,  0.00000011],
           [ 0.00000089,  0.00000077],
           [ 0.00000021,  0.00000091]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('l', 'floating', range_='[{};inf)'.format(_min_l))
        _numeric('w', 'floating', range_='[{};inf)'.format(_min_w))

        if l != w:
            msg = ('The value of >>h<<, {!r}, must equal the value of >>w<<, '
                   '{!r}, for an archimedian spiral to make sense.')
            raise ValueError(msg.format(l, w))

        _numeric('speed', 'floating', range_='[{};inf)'.format(_min_speed))
        _numeric('sample_rate', 'floating',
                 range_='[{};inf)'.format(_min_sample_rate))
        _numeric('time', 'floating', range_='[{};inf)'.format(_min_time))
        _numeric('rect_area', 'boolean')

    validate_input()

    sample_period = 1 / sample_rate
    r_end = np.min([l, w]) / 2

    if rect_area:
        # Sample the "corners" of a rectangular area
        r_end *= np.sqrt(2)

    pitch = np.pi * r_end ** 2 / (speed * time)

    # Starting at t=0 yields a divide-by-zero problem. Therefore we start the
    # spiral at t=sample-period
    sample_times = np.linspace(sample_period, time, sample_rate * time - 1)

    r = np.sqrt(pitch * speed / np.pi * sample_times)
    theta = np.sqrt(4 * np.pi * speed * sample_times / pitch)

    coords = np.zeros((len(sample_times), 2))
    coords[:, 0] = r * np.cos(theta)
    coords[:, 1] = r * np.sin(theta)

    # FIXME: This simple stretch yields a non-constant linear speed, which
    # clashes with the idea of the current interface to all sample methods.
    """
    if w > l:
        coords[:, 0] = coords[:, 0] * w / l
    else:
        coords[:, 1] = coords[:, 1] * l / w
    """

    if rect_area:
        # We assume the tip follows the spiral outside of the rectangular area
        # These coordinates should not be included in the sampled area
        within_rect_area = np.where(
            np.all(np.abs(coords) <= (w / 2, l / 2), axis=1))
        coords = coords[within_rect_area]

    coords = coords + np.array([w / 2, l / 2])

    return coords



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests for `magni.cs.phase_transition`.

Routine listings
----------------
TestAnalysis(unittest.TestCase)
    Tests of _analysis.run
TestAnalysisEstimatePT(unittest.TestCase)
    Tests of the _analysis._estimatePT function.
TestDataGenerateMatrix(unittest.TestCase)
    Tests of the _data.generate_matrix function.
TestDataGenerateNoise(unittest.TestCase)
    Tests of the _data.generate_noise function.
TestDataGenerateVector(unittest.TestCase)
    Tests of the _data.generate_vector function.
TestDetermine(unittest.TestCase)
    Tests of _util.determine function.
class TestIO(unittest.TestCase)
    Tests of the io module.
TestSimulationSimulate(unittest.TestCase)
    Tests of the _simulation._simulate function.

"""

from __future__ import division
from contextlib import contextmanager
import os
import shutil
import sys
import unittest
import warnings

import numpy as np
import tables as tb
try:
    import sklearn
except ImportError:
    sklearn = False

import magni


class TestAnalysis(unittest.TestCase):
    """
    Tests of _analysis.run

    **Testing Strategy**
    Test logistic regression phase transition fit against reference.

    """

    def setUp(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        self.file_ = 'pt_test_data.hdf5'
        shutil.copy(os.path.join(self.path, self.file_), self.file_)
        self.label = 'test_pt_data'
        magni.cs.phase_transition.config.update(
            {'delta': np.linspace(0.025, 1.00, 40).tolist(),
             'monte_carlo': 10,
             'problem_size': 32**2,
             'rho': np.linspace(0.01, 1.00, 100).tolist(),
             'seed': 6021})

        self.assertTrue(os.path.exists(self.file_))
        with tb.File(self.file_, mode='r') as h5_file:
            self.reference_pt = h5_file.root.ref_phase_transition.read()
            self.reference_pt_80db = (
                h5_file.root.ref_phase_transition_highsens.read())

        self.synthetic_file = 'cs_phase_transition_analysis_test.hdf5'
        self.assertFalse(os.path.exists(self.synthetic_file))
        with tb.File(self.synthetic_file, mode='a') as h5_file:
            h5_file.create_group('/', 'all_or_none_pt')
            h5_file.create_array(
                '/all_or_none_pt', 'dist',
                np.concatenate(
                    (np.zeros((1, 100, 10)), np.ones((1, 100, 10))), axis=0))

    def tearDown(self):
        magni.cs.phase_transition.config.reset()
        if os.path.exists(self.file_):
            os.remove(self.file_)
        self.assertFalse(os.path.exists(self.file_))

        if os.path.exists(self.synthetic_file):
            os.remove(self.synthetic_file)
        self.assertFalse(os.path.exists(self.synthetic_file))

    def test_built_in_solver_reference_default(self):
        self.assertEqual(
            magni.cs.phase_transition.config['logit_solver'], 'built-in')
        self.assertEqual(magni.cs.phase_transition.config['SNR'], 40)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            magni.cs.phase_transition._analysis.run(self.file_, self.label)
        with tb.File(self.file_, mode='r') as h5_file:
            estimated_pt = h5_file.get_node(
                '/'.join(['', self.label, 'phase_transition'])).read()
            estimated_percentiles = h5_file.get_node(
                '/'.join(['', self.label, 'phase_transition_percentiles'])
            ).read()
        self.assertTrue(np.allclose(estimated_pt, self.reference_pt))
        for p in range(estimated_percentiles.shape[0]):
            if p < 2:
                # 10%, 25% percentiles - should be above 50%
                self.assertTrue(
                    np.all(estimated_percentiles[p, :] >= self.reference_pt))
            else:
                # 75%, 90% percentiles - should be below 50%
                self.assertTrue(
                    np.all(estimated_percentiles[p, :] <= self.reference_pt))

    def test_built_in_solver_reference_80dB(self):
        magni.cs.phase_transition.config['SNR'] = 80
        self.assertEqual(
            magni.cs.phase_transition.config['logit_solver'], 'built-in')
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            magni.cs.phase_transition._analysis.run(self.file_, self.label)
        with tb.File(self.file_, mode='r') as h5_file:
            estimated_pt = h5_file.get_node(
                '/'.join(['', self.label, 'phase_transition'])).read()
            estimated_percentiles = h5_file.get_node(
                '/'.join(['', self.label, 'phase_transition_percentiles'])
            ).read()
        self.assertTrue(
            np.allclose(estimated_pt, self.reference_pt_80db, atol=1e-1))
        for p in range(estimated_percentiles.shape[0]):
            if p < 2:
                # 10%, 25% percentiles - should be above 50%
                self.assertTrue(
                    np.all(
                        estimated_percentiles[p, :] >= self.reference_pt_80db))
            else:
                # 75%, 90% percentiles - should be below 50%
                self.assertTrue(
                    np.all(
                        estimated_percentiles[p, :] <= self.reference_pt_80db))
        self.assertEqual(magni.cs.phase_transition.config['SNR'], 80)

    @unittest.skipIf(not sklearn, 'Scikit-learn is not available')
    def test_sklearn_solver_reference_default(self):
        magni.cs.phase_transition.config['logit_solver'] = 'sklearn'
        self.assertEqual(magni.cs.phase_transition.config['SNR'], 40)
        magni.cs.phase_transition._analysis.run(self.file_, self.label)
        with tb.File(self.file_, mode='r') as h5_file:
            estimated_pt = h5_file.get_node(
                '/'.join(['', self.label, 'phase_transition'])).read()
            estimated_percentiles = h5_file.get_node(
                '/'.join(['', self.label, 'phase_transition_percentiles'])
            ).read()
        self.assertTrue(
            np.allclose(estimated_pt, self.reference_pt, atol=1e-1))
        for p in range(estimated_percentiles.shape[0]):
            if p < 2:
                # 10%, 25% percentiles - should be above 50%
                self.assertTrue(np.all(estimated_percentiles[p, :-1] >=
                                       self.reference_pt[:-1]))
            else:
                # 75%, 90% percentiles - should be below 50%
                self.assertTrue(
                    np.all(estimated_percentiles[p, :] <= self.reference_pt))
        self.assertEqual(
            magni.cs.phase_transition.config['logit_solver'], 'sklearn')

    @unittest.skipIf(not sklearn, 'Scikit-learn is not available')
    def test_sklearn_solver_reference_80dB(self):
        magni.cs.phase_transition.config['SNR'] = 80
        magni.cs.phase_transition.config['logit_solver'] = 'sklearn'
        magni.cs.phase_transition._analysis.run(self.file_, self.label)
        with tb.File(self.file_, mode='r') as h5_file:
            estimated_pt = h5_file.get_node(
                '/'.join(['', self.label, 'phase_transition'])).read()
            estimated_percentiles = h5_file.get_node(
                '/'.join(['', self.label, 'phase_transition_percentiles'])
            ).read()
        self.assertTrue(
            np.allclose(estimated_pt, self.reference_pt_80db))
        for p in range(estimated_percentiles.shape[0]):
            if p < 2:
                # 10%, 25% percentiles - should be above 50%
                self.assertTrue(np.all(estimated_percentiles[p, :-1] >=
                                       self.reference_pt_80db[:-1]))
            else:
                # 75%, 90% percentiles - should be below 50%
                self.assertTrue(
                    np.all(
                        estimated_percentiles[p, :] <= self.reference_pt_80db))
        self.assertEqual(
            magni.cs.phase_transition.config['logit_solver'], 'sklearn')
        self.assertEqual(magni.cs.phase_transition.config['SNR'], 80)

    def test_built_in_solver_all_or_none(self):
        magni.cs.phase_transition.config['delta'] = [0.2, 0.3]
        self.assertEqual(
            magni.cs.phase_transition.config['logit_solver'], 'built-in')
        magni.cs.phase_transition._analysis.run(
            self.synthetic_file, 'all_or_none_pt')
        with tb.File(self.synthetic_file, mode='r') as h5_file:
            estimated_pt = h5_file.get_node(
                '/'.join(['', 'all_or_none_pt', 'phase_transition'])).read()
        self.assertTrue(np.allclose(estimated_pt, np.array([1, 0])))

    @unittest.skipIf(not sklearn, 'Scikit-learn is not available')
    def test_sklearn_solver_all_or_none(self):
        magni.cs.phase_transition.config['delta'] = [0.2, 0.3]
        magni.cs.phase_transition.config['logit_solver'] = 'sklearn'
        self.assertEqual(
            magni.cs.phase_transition.config['logit_solver'], 'sklearn')
        magni.cs.phase_transition._analysis.run(
            self.synthetic_file, 'all_or_none_pt')
        with tb.File(self.synthetic_file, mode='r') as h5_file:
            estimated_pt = h5_file.get_node(
                '/'.join(['', 'all_or_none_pt', 'phase_transition'])
            ).read()
        self.assertTrue(np.allclose(estimated_pt, np.array([1, 0])))


class TestAnalysisEstimatePT(unittest.TestCase):
    """
    Tests of the _analysis._estimatePT function.

    **Testing Strategy**
    Test af standard logistic fit and its "reverse".

    """

    def test_std_logistic(self):
        rho = np.linspace(0, 1, 11)
        success = np.array([1., 1., 1., 1., 1., 0., 0., 0., 0., 0., 0.])

        pt_val = magni.cs.phase_transition._analysis._estimate_PT(
            rho, success, [0.5])
        self.assertEqual(np.round(pt_val[0], 2), 0.45)

    def test_non_convergence(self):
        rho = np.linspace(0, 1, 11)
        success = np.array([1., 1., 1., 1., 1., 0., 0., 0., 0., 0., 0.])[::-1]

        with self.assertRaises(RuntimeWarning):
            pt_val = magni.cs.phase_transition._analysis._estimate_PT(
                rho, success, [0.5])


class TestBackup(unittest.TestCase):
    """
    Test of the _backup functionality.

    **Testing Strategy**
    Create a backup and test get and set methods.

    """

    def setUp(self):
        self.path = 'cs_phase_transition_backup_tests.hdf5'
        self.assertFalse(os.path.exists(self.path))
        magni.cs.phase_transition._backup.create(self.path)
        self.assertTrue(os.path.exists(self.path))

    def tearDown(self):
        if os.path.exists(self.path):
            os.remove(self.path)
        self.assertFalse(os.path.exists(self.path))

    def test_get_backup(self):
        done = magni.cs.phase_transition._backup.get(self.path)
        self.assertFalse(np.any(done))

    def test_set_backup(self):
        magni.cs.phase_transition._backup.set(
            self.path, (0, 0), 1.0, 2.0, 3.0, 4.0)
        done = magni.cs.phase_transition._backup.get(self.path)
        self.assertTrue(np.any(done))

        with tb.File(self.path, mode='r') as h5:
            self.assertEqual(h5.root.time[0, 0], 1.0)
            self.assertEqual(h5.root.dist[0, 0], 2.0)
            self.assertEqual(h5.root.mse[0, 0], 3.0)
            self.assertEqual(h5.root.norm[0, 0], 4.0)


class TestDataGenerateMatrix(unittest.TestCase):
    """
    Tests of the _data.generate_matrix function.

    **Testing Strategy**
    Matrices generated by the function are compared to known "true" matrices.

    """

    def setUp(self):
        self.l = 4
        self.m = 3
        self.n = 9
        self.seed = 6021
        np.random.seed(self.seed)

    def tearDown(self):
        magni.cs.phase_transition.config.reset()

    def test_default(self):
        A_true = np.array([[0.91579356, 0.8554431, -0.73516709, -0.5351897,
                            0.92666919, 0.73139786,  0.36218319, -0.59300362,
                            0.90664675],
                           [-0.0671974, -0.47660808, -0.66815151, -0.72081901,
                            -0.34468206, -0.47720612, -0.16854207, -0.36796994,
                            0.39610724],
                           [-0.39598822, -0.20263723, -0.11446798, -0.44044516,
                            -0.14992827, -0.4871668, -0.91674255, -0.71620166,
                            0.14522645]])

        A = magni.cs.phase_transition._data.generate_matrix(self.m, self.n)

        self.assertTrue(np.allclose(A, A_true))

    def test_uniform_line_2d_dct(self):
        Phi_true = np.array([[1., 0., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 1., 0., 0., 0., 0., 0., 0., 0.],
                             [0., 0., 0., 0., 0., 0., 0., 0., 1.]])

        n_sqrt = np.sqrt(self.n)
        self.assertTrue(n_sqrt.is_integer())
        n_sqrt = int(n_sqrt)
        Psi_true = magni.imaging.dictionaries.get_DCT((n_sqrt, n_sqrt)).A

        A_true = Phi_true.dot(Psi_true)

        magni.cs.phase_transition.config['system_matrix'] = 'RandomDCT2D'
        A = magni.cs.phase_transition._data.generate_matrix(self.m, self.n).A

        self.assertEqual(magni.cs.phase_transition.config['system_matrix'],
                         'RandomDCT2D')
        self.assertTrue(np.allclose(A, A_true))

        with self.assertRaises(ValueError):
            magni.cs.phase_transition._data.generate_matrix(3, 7)

    def test_custom_system_matrix(self):
        magni.cs.phase_transition.config['system_matrix'] = 'custom'
        with self.assertRaises(TypeError):
            # One must configure a system matrix factory
            magni.cs.phase_transition._data.generate_matrix(self.m, self.n)

        def generate_custom_ndarray(m, n):
            return np.random.randint(10, size=(m, n))

        magni.cs.phase_transition.config[
            'custom_system_matrix_factory'] = generate_custom_ndarray
        A_true_ndarray = np. array([[3, 7, 6, 5, 4, 2, 2, 3, 6],
                                    [3, 9, 8, 7, 0, 1, 4, 0, 5],
                                    [2, 7, 6, 0, 9, 2, 4, 5, 5]])
        A = magni.cs.phase_transition._data.generate_matrix(self.m, self.n)
        self.assertTrue(np.allclose(A, A_true_ndarray))

        def generate_wrong_ndarray(m, n):
            return np.random.randint(10, size=(n, m))

        magni.cs.phase_transition.config[
            'custom_system_matrix_factory'] = generate_wrong_ndarray
        with self.assertRaises(ValueError):
            magni.cs.phase_transition._data.generate_matrix(self.m, self.n)

        self.assertEqual(
            magni.cs.phase_transition.config['system_matrix'], 'custom')


class TestDataGenerateNoise(unittest.TestCase):
    """
    Tests of the _data.generate_noise function.

    **Testing Strategy**
    Various noise vectors are generated from a specified SNR. An estimated SNR
    is compared to the specified SNR.

    """

    def setUp(self):
        np.random.seed(6021)
        self.n = 1241415
        self.m = 148411
        self.k = 12151
        self.SNRs = [float(SNR) if SNR % 2 == 0 else SNR
                     for SNR in np.random.choice(
                             np.arange(1, 60), size=10, replace=False)]
        self.coefficients = ['rademacher', 'gaussian', 'laplace', 'bernoulli']

    def tearDown(self):
        magni.cs.phase_transition.config.reset()

    def test_AWGN(self):
        magni.cs.phase_transition.config['noise'] = 'AWGN'
        for coeffs in self.coefficients:
            magni.cs.phase_transition.config['coefficients'] = coeffs
            signal_vector = magni.cs.phase_transition._data.generate_vector(
                self.n, self.k)
            estimated_signal_power = (
                1.0/self.n * np.linalg.norm(signal_vector)**2)

            for SNR in self.SNRs:
                magni.cs.phase_transition.config['SNR'] = SNR
                noise_vector = magni.cs.phase_transition._data.generate_noise(
                    self.m, self.n, self.k)
                estimated_noise_power = (
                    1.0/self.m * np.linalg.norm(noise_vector)**2)
                estimated_SNR = 10 * np.log10(estimated_signal_power /
                                              estimated_noise_power)
                self.assertTrue(np.allclose(estimated_SNR, SNR, atol=5e-1))
                self.assertEqual(magni.cs.phase_transition.config['SNR'], SNR)

            self.assertEqual(
                magni.cs.phase_transition.config['coefficients'], coeffs)

        self.assertEqual(magni.cs.phase_transition.config['noise'], 'AWGN')

    def test_AWLN(self):
        magni.cs.phase_transition.config['noise'] = 'AWLN'
        for coeffs in self.coefficients:
            magni.cs.phase_transition.config['coefficients'] = coeffs
            signal_vector = magni.cs.phase_transition._data.generate_vector(
                self.n, self.k)
            estimated_signal_power = (
                1.0/self.n * np.linalg.norm(signal_vector)**2)

            for SNR in self.SNRs:
                magni.cs.phase_transition.config['SNR'] = SNR
                noise_vector = magni.cs.phase_transition._data.generate_noise(
                    self.m, self.n, self.k)
                estimated_noise_power = (
                    1.0/self.m * np.linalg.norm(noise_vector)**2)
                estimated_SNR = 10 * np.log10(estimated_signal_power /
                                              estimated_noise_power)

                self.assertTrue(np.allclose(estimated_SNR, SNR, atol=1e-1))
                self.assertEqual(magni.cs.phase_transition.config['SNR'], SNR)

            self.assertEqual(
                magni.cs.phase_transition.config['coefficients'], coeffs)

        self.assertEqual(magni.cs.phase_transition.config['noise'], 'AWLN')

    def test_custom_noise(self):
        magni.cs.phase_transition.config['noise'] = 'custom'
        with self.assertRaises(TypeError):
            # One must configure a noise factory
            magni.cs.phase_transition._data.generate_noise(
                self.m, self.n, self.k)

        def generate_custom_noise(m, n, k, noise_power):
            return np.random.normal(scale=np.sqrt(noise_power), size=(m, 1))

        magni.cs.phase_transition.config[
            'custom_noise_factory'] = generate_custom_noise
        for coeffs in self.coefficients:
            magni.cs.phase_transition.config['coefficients'] = coeffs
            signal_vector = magni.cs.phase_transition._data.generate_vector(
                self.n, self.k)
            estimated_signal_power = (
                1.0/self.n * np.linalg.norm(signal_vector)**2)

            for SNR in self.SNRs:
                magni.cs.phase_transition.config['SNR'] = SNR
                noise_vector = magni.cs.phase_transition._data.generate_noise(
                    self.m, self.n, self.k)
                estimated_noise_power = (
                    1.0/self.m * np.linalg.norm(noise_vector)**2)
                estimated_SNR = 10 * np.log10(estimated_signal_power /
                                              estimated_noise_power)

                print(estimated_SNR, SNR)

                self.assertTrue(np.allclose(estimated_SNR, SNR, atol=1e-1))
                self.assertEqual(magni.cs.phase_transition.config['SNR'], SNR)

            self.assertEqual(
                magni.cs.phase_transition.config['coefficients'], coeffs)

        def generate_wrong_noise_shape(m, n, k, noise_power):
            return np.random.normal(scale=np.sqrt(noise_power), size=(1, n))

        magni.cs.phase_transition.config[
            'custom_noise_factory'] = generate_wrong_noise_shape
        with self.assertRaises(ValueError):
            magni.cs.phase_transition._data.generate_noise(
                self.m, self.n, self.k)

        self.assertEqual(magni.cs.phase_transition.config['noise'], 'custom')

    def test_None_noise(self):
        self.assertEqual(magni.cs.phase_transition.config['noise'], None)
        with self.assertRaises(RuntimeError):
            magni.cs.phase_transition._data.generate_noise(
                self.m, self.n, self.k)


class TestDataGenerateVector(unittest.TestCase):
    """
    Tests of the _data.generate_vector function.

    **Testing Strategy**
    Vectors generated by the function are compared to known "true" vectors.

    """

    def setUp(self):
        self.n = 17
        self.k = 8
        self.seed = 6021
        np.random.seed(self.seed)

    def tearDown(self):
        magni.cs.phase_transition.config.reset()

    def test_default(self):
        x_true = np.array([[1.], [1.], [1.], [-1.], [-1.], [1.], [-1.], [-1.],
                           [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.]
                           ])

        x = magni.cs.phase_transition._data.generate_vector(self.n, self.k)

        self.assertTrue(np.allclose(x_true, x))

    def test_rademacher(self):
        magni.cs.phase_transition.config.update(coefficients='rademacher')
        x_true = np.array([[1.], [1.], [1.], [-1.], [-1.], [1.], [-1.], [-1.],
                           [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.]
                           ])

        x = magni.cs.phase_transition._data.generate_vector(self.n, self.k)

        self.assertTrue(np.allclose(x_true, x))
        self.assertEqual(
            magni.cs.phase_transition.config['coefficients'], 'rademacher')

    def test_gaussian(self):
        magni.cs.phase_transition.config.update(coefficients='gaussian')
        x_true = np.array([[0.97267375], [0.95762797], [-1.26236611],
                           [-0.71956614], [1.96767297], [0.5739882],
                           [0.49924214], [-0.98133296],
                           [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.]
                           ])

        x = magni.cs.phase_transition._data.generate_vector(self.n, self.k)

        self.assertTrue(np.allclose(x_true, x))
        self.assertEqual(
            magni.cs.phase_transition.config['coefficients'], 'gaussian')

    def test_laplace(self):
        magni.cs.phase_transition.config.update(coefficients='laplace')
        x_true = np.array([[0.86206783], [-1.74301165], [0.58042009],
                           [0.59285876], [-0.34547475], [-0.71840635],
                           [1.52441889], [1.24356796],
                           [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.]
                           ])

        x = magni.cs.phase_transition._data.generate_vector(self.n, self.k)

        self.assertTrue(np.allclose(x_true, x))
        self.assertEqual(
            magni.cs.phase_transition.config['coefficients'], 'laplace')

    def test_bernoulli(self):
        magni.cs.phase_transition.config.update(coefficients='bernoulli')
        x_true = np.array([[1.], [1.], [1.], [1.], [1.], [1.], [1.], [1.],
                           [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.]
                           ])

        x = magni.cs.phase_transition._data.generate_vector(self.n, self.k)

        self.assertTrue(np.allclose(x_true, x))
        self.assertEqual(
            magni.cs.phase_transition.config['coefficients'], 'bernoulli')

    def test_linear_support(self):
        support_distribution = np.reshape((np.arange(self.n) + 1) /
                                          (self.n * (self.n + 1) / 2),
                                          (self.n, 1))
        magni.cs.phase_transition.config.update(
            support_distribution=support_distribution)
        x_true = np.array([[0.], [0.], [0.], [0.], [-1.], [0.], [0.], [0.],
                           [-1.], [1.], [0.], [-1.], [1.], [0.], [-1.], [1.],
                           [-1.]])

        x = magni.cs.phase_transition._data.generate_vector(self.n, self.k)

        self.assertTrue(np.allclose(x_true, x))
        self.assertEqual(
            magni.cs.phase_transition.config['coefficients'],
            'rademacher')
        self.assertIs(
            magni.cs.phase_transition.config['support_distribution'],
            support_distribution)

    def test_exponential_support(self):
        support_distribution = np.exp(-np.arange(self.n)).reshape(self.n, 1)
        support_distribution /= np.sum(support_distribution)
        magni.cs.phase_transition.config.update(
            support_distribution=support_distribution)
        x_true = np.array([[-1.], [1.], [-1.], [-1.], [1.], [1.], [1.], [-1.],
                           [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.], [0.]
                           ])

        x = magni.cs.phase_transition._data.generate_vector(self.n, self.k)
        self.assertTrue(np.allclose(x_true, x))
        self.assertEqual(
            magni.cs.phase_transition.config['coefficients'],
            'rademacher')
        self.assertIs(
            magni.cs.phase_transition.config['support_distribution'],
            support_distribution)

    def test_uniform_support(self):
        support_distribution = np.ones((self.n, 1)) / self.n
        magni.cs.phase_transition.config.update(
            support_distribution=support_distribution)
        x_true = np.array([[0.], [-1.], [0.], [0.], [-1.], [0.], [-1.], [0.],
                           [-1.], [0.], [0.], [0.], [1.], [1.], [-1.], [1.],
                           [0.]])

        x = magni.cs.phase_transition._data.generate_vector(self.n, self.k)
        self.assertTrue(np.allclose(x_true, x))
        self.assertEqual(
            magni.cs.phase_transition.config['coefficients'],
            'rademacher')
        self.assertIs(
            magni.cs.phase_transition.config['support_distribution'],
            support_distribution)


class TestDetermine(unittest.TestCase):
    """
    Tests of _util.determine function.

    This function is the main entry function
    `magni.cs.phase_transition.determine`.

    **Testing Strategy**
    A very small phase transition run is compared to a reference to test that
    phase transitions are stored correctly.

    The kwarg pass through and pre_simulation_hook functionality of the phase
    transition subpackage is tested by comparing the reported setup to a
    reference.

    Error handling is tested separately.

    """

    def setUp(self):
        magni.utils.multiprocessing.config['workers'] = 1
        self.determine_test_path = 'cs_phase_transition_determine_test.hdf5'

        magni.cs.reconstruction.gamp.config.update({
            'input_channel_parameters': {
                'tau': 0.5, 'theta_bar': 0, 'theta_tilde': 1, 'use_em': False},
            'output_channel_parameters': {
                'sigma_sq': 1, 'noise_level_estimation': 'sample_variance'},
            'iterations': 10,
            'report_A_asq_setup': True
        })

        magni.cs.phase_transition.config.update({
            'seed': 6021,
            'problem_size': 32**2,
            'delta': [0.3],
            'rho': [0.2, 0.3, 0.4, 0.5, 0.6, 0.8],
            'coefficients': 'laplace',
            'system_matrix': 'RandomDCT2D',
            'monte_carlo': 1
        })

        self.dummy_path = 'cs_phase_transition_error_handling.hdf5'
        self.assertFalse(os.path.exists(self.dummy_path))

    def tearDown(self):
        magni.utils.multiprocessing.config.reset()
        magni.cs.reconstruction.gamp.config.reset()
        magni.cs.phase_transition.config.reset()
        if os.path.exists(self.dummy_path):
            os.remove(self.dummy_path)
        self.assertFalse(os.path.exists(self.dummy_path))

    def test_sum_approximation_gamp(self):
        magni.utils.multiprocessing.config['workers'] = 0
        file_ = 'ma_gamp'
        with _capture_output(file_):
            magni.cs.phase_transition.determine(
                magni.cs.reconstruction.gamp.run,
                'cs_phase_transition_determine_test.hdf5',
                label='ma_gamp_test',
                pre_simulation_hook=_set_tau)

        with open(file_ + '.stdout', mode='r') as stdout:
            for line in stdout:
                if line.startswith('GAMP'):
                    self.assertEqual(
                        line.strip()[:79],
                        'GAMP is using the A_asq: ' +
                        '<magni.utils.matrices.SumApproximationMatrix' +
                        ' object at')
                else:
                    self.assertEqual(
                        line.strip(),
                        'The sum approximation method is: rangan')

    def test_full_gamp(self):
        magni.utils.multiprocessing.config['workers'] = 0
        file_ = 'full_gamp'
        with _capture_output(file_):
            magni.cs.phase_transition.determine(
                magni.cs.reconstruction.gamp.run,
                'cs_phase_transition_determine_test.hdf5',
                label='full_gamp_test',
                pre_simulation_hook=_construct_A_asq_and_set_tau)

        with open(file_ + '.stdout', mode='r') as stdout:
            for line in stdout:
                if line.startswith('GAMP'):
                    self.assertEqual(
                        line.strip()[:73],
                        'GAMP is using the A_asq: ' +
                        '<magni.utils.matrices.MatrixCollection object at')
                else:
                    self.assertEqual(
                        line.strip(),
                        'The sum approximation method is: None')

    def test_z_correct_dists_ma_gamp(self):
        with tb.File('cs_phase_transition_determine_test.hdf5') as h5_file:
            ma_gamp_dist = h5_file.get_node('/ma_gamp_test/dist').read()

        if sys.version_info[0] == 2:
            ma_gamp_dist_ref = np.array(
                [[[0.00052035], [0.00670562], [0.04877081],
                  [0.11715555], [0.17149385], [0.41228690]]])
        else:
            # random.sample is not stable across Python 2 --> 3 border
            ma_gamp_dist_ref = np.array(
                [[[2.72592311e-04], [2.48630874e-03], [6.26038548e-02],
                  [1.65474043e-01], [3.03143840e-01], [3.92334845e-01]]])

        self.assertTrue(np.allclose(ma_gamp_dist, ma_gamp_dist_ref))

    def test_z_correct_dists_full_gamp(self):
        with tb.File('cs_phase_transition_determine_test.hdf5') as h5_file:
            full_gamp_dist = h5_file.get_node('/full_gamp_test/dist').read()

        if sys.version_info[0] == 2:
            full_gamp_dist_ref = np.array(
                [[[3.72378090e-04], [4.89454444e-03], [4.51150168e-02],
                  [1.11151619e-01], [1.70156455e-01], [4.09506352e-01]]])
        else:
            # random.sample is not stable across Python 2 --> 3 border
            full_gamp_dist_ref = np.array(
                [[[1.62446873e-04], [2.36392118e-03], [6.09754113e-02],
                  [1.61375617e-01], [2.92454219e-01], [3.91102236e-01]]])

        self.assertTrue(np.allclose(full_gamp_dist, full_gamp_dist_ref))

    def test_invalid_label_character(self):
        with self.assertRaises(RuntimeError):
            magni.cs.phase_transition.determine(
                _dummy_func, self.dummy_path, label='(')

    def test_invalid_label_path(self):
        with self.assertRaises(RuntimeError):
            magni.cs.phase_transition.determine(
                _dummy_func, self.dummy_path, label='/some/path/')

    def test_noncallable_pre_simulation_hook(self):
        with self.assertRaises(RuntimeError):
            magni.cs.phase_transition.determine(
                _dummy_func, self.dummy_path, pre_simulation_hook='')

    def test_overwrite_label(self):
        magni.utils.multiprocessing.config['workers'] = 0
        label = 'test'
        with tb.File(self.dummy_path, mode='w') as h5:
            h5.create_array('/', label, np.array([1]))

        with self.assertRaises(IOError):
            magni.cs.phase_transition.determine(
                _dummy_func, self.dummy_path, label=label)

        with self.assertRaises(RuntimeError):
            # Should overwrite label but fail on executing _dummy_func
            magni.cs.phase_transition.determine(
                _dummy_func, self.dummy_path, label=label, overwrite=True)


class TestIO(unittest.TestCase):
    """
    Tests of the io module.

    **Testing Strategy**
    Input/Output values are compared to a reference.

    """

    def setUp(self):
        self.path = os.path.dirname(os.path.abspath(__file__))
        self.file_ = 'pt_test_data.hdf5'
        shutil.copy(os.path.join(self.path, self.file_), self.file_)
        self.assertTrue(os.path.exists(self.file_))

        self.label = 'test_io_data'
        with tb.File(self.file_, mode='r') as h5_file:
            self.reference_rho = h5_file.get_node(
                '/' + self.label + '/phase_transition').read()
        self.default_delta = np.linspace(0, 1, len(self.reference_rho) + 1)[1:]

    def tearDown(self):
        if os.path.exists(self.file_):
            os.remove(self.file_)
        self.assertFalse(os.path.exists(self.file_))

    def test_default(self):
        delta, rho = magni.cs.phase_transition.io.load_phase_transition(
            self.file_, label=self.label)
        self.assertTrue(np.allclose(delta, self.default_delta))
        self.assertTrue(np.allclose(rho, self.reference_rho))

    def test_custom_delta(self):
        custom_delta = np.ones(self.reference_rho.shape)
        delta, rho = magni.cs.phase_transition.io.load_phase_transition(
            self.file_, label=self.label, delta=custom_delta)
        self.assertTrue(np.allclose(delta, custom_delta))
        self.assertTrue(np.allclose(rho, self.reference_rho))

    def test_invalid_delta(self):
        with self.assertRaises(ValueError):
            # Invalid delta.shape
            magni.cs.phase_transition.io.load_phase_transition(
                self.file_, label=self.label, delta=np.zeros(1))

        with self.assertRaises(ValueError):
            # Invalid range of delta
            magni.cs.phase_transition.io.load_phase_transition(
                self.file_, label=self.label,
                delta=3 * np.ones(self.reference_rho.shape))


class TestSimulationSimulate(unittest.TestCase):
    """
    Tests of the _simulation._simulate function.

    **Testing Strategy**
    Phase transition distance simulations are compared to known "true"
    distances.

    """

    _backup = magni.cs.phase_transition._simulation._backup

    def setUp(self):
        self.seed = 6021
        self.delta = [0.2]
        self.rho = [0.1]
        self.monte_carlo = 1
        self.coefficients = 'rademacher'
        self.n = magni.cs.phase_transition.config['problem_size']

        class BackupStub():
            def set(self, path, ij_tup, stat_time, stat_dist, stat_mse,
                    stat_norm):
                self.path = path
                self.ij_tup = ij_tup
                self.stat_time = stat_time
                self.stat_dist = stat_dist
                self.stat_mse = stat_mse
                self.stat_norm = stat_norm

        self.backup_stub = BackupStub()

        magni.cs.phase_transition._simulation._backup = self.backup_stub

    def tearDown(self):
        magni.cs.reconstruction.it.config.reset()
        magni.cs.phase_transition.config.reset()
        magni.cs.phase_transition._simulation._backup = self.__class__._backup

    def test_default(self):
        magni.cs.phase_transition.config.update(
            {'seed': self.seed, 'delta': self.delta, 'rho': self.rho,
             'monte_carlo': self.monte_carlo,
             'coefficients': self.coefficients})

        path = 'path'
        ij_tup = (0, 0)

        self.assertFalse(self.backup_stub.__dict__)

        magni.cs.phase_transition._simulation._simulate(
            magni.cs.reconstruction.it.run, ij_tup, [self.seed], path)

        self.assertEqual(self.backup_stub.path, path)
        self.assertEqual(self.backup_stub.ij_tup, ij_tup)
        self.assertTrue(np.allclose(self.backup_stub.stat_mse,
                                    np.array([[1.08899121e-07]])))
        self.assertTrue(np.allclose(self.backup_stub.stat_norm,
                                    np.array([4])))
        self.assertTrue(np.allclose(
            self.backup_stub.stat_dist,
            self.backup_stub.stat_mse * self.n / self.backup_stub.stat_norm**2)
        )

    def test_pre_sim_hook(self):
        magni.cs.phase_transition.config.update(
            {'seed': self.seed, 'delta': self.delta, 'rho': self.rho,
             'monte_carlo': self.monte_carlo,
             'coefficients': self.coefficients})

        def pre_sim_hook(var):
            magni.cs.reconstruction.it.config.update(
                {'threshold': 'fixed',
                 'threshold_fixed': int(self.rho[0] * self.delta[0] * self.n),
                 'kappa': 'fixed', 'kappa_fixed': 0.45, 'tolerance': 1e-4})

        path = 'path'
        ij_tup = (0, 0)

        self.assertFalse(self.backup_stub.__dict__)

        magni.cs.phase_transition._simulation._simulate(
            magni.cs.reconstruction.it.run, ij_tup, [self.seed], path,
            pre_sim_hook)

        self.assertEqual(self.backup_stub.path, path)
        self.assertEqual(self.backup_stub.ij_tup, ij_tup)
        self.assertTrue(np.allclose(self.backup_stub.stat_mse,
                                    np.array([[2.47979301e-10]])))
        self.assertTrue(np.allclose(self.backup_stub.stat_norm,
                                    np.array([4])))
        self.assertTrue(np.allclose(
            self.backup_stub.stat_dist,
            self.backup_stub.stat_mse * self.n / self.backup_stub.stat_norm**2)
        )

    def test_true_linear_support(self):
        support_distribution = np.reshape((np.arange(self.n) + 1) /
                                          (self.n * (self.n + 1) / 2),
                                          (self.n, 1))
        weights = np.linspace(0.5, 1.0, self.n).reshape(-1, 1)
        magni.cs.phase_transition.config.update(
            {'seed': self.seed, 'delta': self.delta, 'rho': self.rho,
             'monte_carlo': self.monte_carlo,
             'coefficients': self.coefficients,
             'support_distribution': support_distribution})
        magni.cs.reconstruction.it.config.update(
            {'threshold': 'fixed',
             'threshold_fixed': int(self.rho[0] * self.delta[0] * self.n),
             'threshold_operator': 'weighted_hard',
             'threshold_weights': weights,
             'kappa': 'fixed', 'kappa_fixed': 0.45, 'tolerance': 1e-4})

        path = 'path'
        ij_tup = (0, 0)

        self.assertFalse(self.backup_stub.__dict__)

        magni.cs.phase_transition._simulation._simulate(
            magni.cs.reconstruction.it.run, ij_tup, [self.seed], path)

        self.assertEqual(self.backup_stub.path, path)
        self.assertEqual(self.backup_stub.ij_tup, ij_tup)
        self.assertTrue(np.allclose(self.backup_stub.stat_mse,
                                    np.array([[3.76042577e-10]])))
        self.assertTrue(np.allclose(self.backup_stub.stat_norm,
                                    np.array([4])))
        self.assertTrue(np.allclose(
            self.backup_stub.stat_dist,
            self.backup_stub.stat_mse * self.n / self.backup_stub.stat_norm**2)
        )

    def test_false_linear_support(self):
        support_distribution = np.reshape((np.arange(self.n) + 1) /
                                          (self.n * (self.n + 1) / 2),
                                          (self.n, 1))
        weights = np.linspace(1.0, 0.5, self.n).reshape(-1, 1)
        magni.cs.phase_transition.config.update(
            {'seed': self.seed, 'delta': self.delta, 'rho': self.rho,
             'monte_carlo': self.monte_carlo,
             'coefficients': self.coefficients,
             'support_distribution': support_distribution})
        magni.cs.reconstruction.it.config.update(
            {'threshold': 'fixed',
             'threshold_fixed': int(self.rho[0] * self.delta[0] * self.n),
             'threshold_operator': 'weighted_hard',
             'threshold_weights': weights,
             'kappa': 'fixed', 'kappa_fixed': 0.45, 'tolerance': 1e-4})

        path = 'path'
        ij_tup = (0, 0)

        self.assertFalse(self.backup_stub.__dict__)

        magni.cs.phase_transition._simulation._simulate(
            magni.cs.reconstruction.it.run, ij_tup, [self.seed], path)

        self.assertEqual(self.backup_stub.path, path)
        self.assertEqual(self.backup_stub.ij_tup, ij_tup)
        self.assertTrue(np.allclose(self.backup_stub.stat_mse,
                                    np.array([[0.00784895]])))
        self.assertTrue(np.allclose(self.backup_stub.stat_norm,
                                    np.array([4])))
        self.assertTrue(np.allclose(
            self.backup_stub.stat_dist,
            self.backup_stub.stat_mse * self.n / self.backup_stub.stat_norm**2)
        )

    def test_noisy_measurements(self):
        magni.cs.phase_transition.config.update(
            {'seed': self.seed, 'delta': self.delta, 'rho': self.rho,
             'monte_carlo': self.monte_carlo,
             'coefficients': self.coefficients,
             'noise': 'AWGN'})

        path = 'path'
        ij_tup = (0, 0)

        self.assertFalse(self.backup_stub.__dict__)

        magni.cs.phase_transition._simulation._simulate(
            magni.cs.reconstruction.it.run, ij_tup, [self.seed], path)

        self.assertEqual(magni.cs.phase_transition.config['noise'], 'AWGN')
        self.assertEqual(self.backup_stub.path, path)
        self.assertEqual(self.backup_stub.ij_tup, ij_tup)
        self.assertTrue(np.allclose(self.backup_stub.stat_mse,
                                    np.array([3.23953658e-07])))
        self.assertTrue(np.allclose(self.backup_stub.stat_norm,
                                    np.array([4])))
        self.assertTrue(np.allclose(
            self.backup_stub.stat_dist,
            self.backup_stub.stat_mse * self.n / self.backup_stub.stat_norm**2)
        )


# Pre-simulation hooks for GAMP
def _construct_A_asq_and_set_tau(var):
    """
    Build the entrywise absolute squared system matrix for AMP.

    Also call `set_tau`, to set the signal density.

    """

    Phi = var['A']._matrices[0]

    dct_mtx = magni.imaging.dictionaries.get_DCT_transform_matrix(
        int(np.sqrt(Phi.shape[1])))
    dct_mtx_sq = dct_mtx**2
    Psi_sq = magni.utils.matrices.Separable2DTransform(
        dct_mtx_sq.T, dct_mtx_sq.T)
    A_asq = magni.utils.matrices.MatrixCollection((Phi, Psi_sq))

    magni.cs.phase_transition.config['algorithm_kwargs'] = {'A_asq': A_asq}
    _set_tau(var)


def _set_tau(var):
    """
    Set GB prior "density" for AMP.

    """

    magni.cs.reconstruction.gamp.config[
        'input_channel_parameters']['tau'] = var['k'] / var['n']


def _dummy_func():
    pass


# Utilities
@contextmanager
def _capture_output(file_):
    """
    Contextmanager for capturing stdout and stderr.

    Inspired by http://stackoverflow.com/a/17981937.

    """

    cur_out, cur_err = sys.stdout, sys.stderr

    try:
        # Apparently, StringIO does not work with multiprocessing in a unittest
        cap_out = open(file_ + '.stdout', mode='a')  # Only capture last write
        cap_err = open(file_ + '.stderr', mode='a')  # Only capture last write
        sys.stdout, sys.stderr = cap_out, cap_err
        yield
    finally:
        sys.stdout, sys.stderr = cur_out, cur_err
        cap_out.close()
        cap_err.close()



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests for `magni.utils.plotting`.

Routine Listings
----------------
TestColourHandling(unittest.TestCase)
    Test of colour handling functionality.
class TestMatplotlibConfiguration(unittest.TestCase)
    Test of matplotlib configuration.

"""

from __future__ import division
import copy
import unittest
import warnings

try:
    from cycler import Cycler
except ImportError:
    Cycler = list
import matplotlib as mpl
import matplotlib.pyplot as plt

import magni


class TestColourHandling(unittest.TestCase):
    """
    Test of colour handling functionality.

    """

    def testColourCollection(self):
        colour_collection = magni.utils.plotting._ColourCollection(
            {'cc1': ((1, 2, 3), (4, 5, 6), (7, 8, 9)),
             'cc2': ((10, 10, 10), (100, 100, 100))})

        true_cc1 = ((0.0039, 0.0078, 0.0118), (0.0157, 0.0196, 0.0235),
                    (0.0275, 0.0314, 0.0353))
        true_cc2 = ((0.0392, 0.0392, 0.0392), (0.3922, 0.3922, 0.3922))
        self.assertEqual(colour_collection['cc1'], true_cc1)
        self.assertEqual(colour_collection['cc2'], true_cc2)


class TestMatplotlibConfiguration(unittest.TestCase):
    """
    Test of matplotlib configuration.

    """

    def setUp(self):
        self.settings_defaults = copy.deepcopy(magni.utils.plotting._settings)
        self.cmap_default = magni.utils.plotting._cmap

    def tearDown(self):
        magni.utils.plotting.setup_matplotlib(
            settings=self.settings_defaults, cmap=self.cmap_default)
        magni.utils.plotting._settings = self.settings_defaults
        magni.utils.plotting._cmap = self.cmap_default
        for name1, settings in self.settings_defaults.items():
            for name2, setting in settings.items():
                self.assertEqual(
                    self._normalise_rcParam(
                        mpl.rcParams['.'.join([name1, name2])]),
                    self._normalise_rcParam(setting))
        self.assertEqual(plt.get_cmap().name, self.cmap_default)

    def test_default_settings(self):
        magni.utils.plotting.setup_matplotlib()
        for name1, settings in self.settings_defaults.items():
            for name2, setting in settings.items():
                self.assertEqual(
                    self._normalise_rcParam(
                        mpl.rcParams['.'.join([name1, name2])]),
                    self._normalise_rcParam(setting))
        self.assertEqual(plt.get_cmap().name, self.cmap_default)

    def test_builtin_cmap(self):
        magni.utils.plotting.setup_matplotlib(cmap='jet')
        self.assertEqual(plt.get_cmap().name, 'jet')

    def test_custom_cmap(self):
        cmap_clone = plt.get_cmap()
        magni.utils.plotting.setup_matplotlib(cmap='jet')
        magni.utils.plotting.setup_matplotlib(
            cmap=('cmap_clone', cmap_clone))
        self.assertEqual(plt.get_cmap(), cmap_clone)

    def test_double_settings_change(self):
        magni.utils.plotting.setup_matplotlib({'font': {'size': 14}})
        magni.utils.plotting.setup_matplotlib({'figure': {'figsize': (1, 1)}})
        self.assertEqual(mpl.rcParams['font.size'], 14)
        self.assertEqual(mpl.rcParams['figure.figsize'], [1, 1])

    def test_invalid_settings(self):
        with warnings.catch_warnings(record=True) as ws:
            warnings.simplefilter('always')
            magni.utils.plotting.setup_matplotlib({'invalid': {'setting': 0}})
        self.assertEqual(len(ws), 1)
        self.assertIsInstance(ws[0].message, UserWarning)
        self.assertEqual(ws[0].message.args[0], "Setting 'invalid' ignored.")

    def _normalise_rcParam(self, rcParam):
        if isinstance(rcParam, (list, Cycler)):
            return tuple(rcParam)
        else:
            return rcParam



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing fast linear operations wrapped in matrix emulators.

Routine listings
----------------
get_DCT(shape, overcomplete_shape=None)
    Get the DCT fast operation dictionary for the given image shape.
get_DFT(shape, overcomplete_shape=None)
    Get the DFT fast operation dictionary for the given image shape.

See Also
--------
magni.imaging.dictionaries._fastops : Fast linear operations.
magni.utils.matrices : Matrix emulators.

"""


from __future__ import division

from pkg_resources import parse_version as _parse_version
from scipy import __version__ as _scipy_version

from magni.imaging.dictionaries import _fastops
from magni.utils.matrices import Matrix as _Matrix
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


def get_DCT(shape, overcomplete_shape=None):
    """
    Get the DCT fast operation dictionary for the given image shape.

    Parameters
    ----------
    shape : list or tuple
        The shape of the image for which the dictionary is the DCT dictionary.
    overcomplete_shape : list or tuple, optional
        The shape of the (overcomplete) frequency domain for the DCT
        dictionary. The entries must be greater than or equal to the
        corresponding entries in `shape`.

    Returns
    -------
    matrix : magni.utils.matrices.Matrix
        The specified DCT dictionary.

    See Also
    --------
    magni.utils.matrices.Matrix : The matrix emulator class.

    Examples
    --------
    Create a dummy image:

    >>> import numpy as np, magni
    >>> img = np.random.randn(64, 64)
    >>> vec = magni.imaging.mat2vec(img)

    Perform DCT in the ordinary way:

    >>> dct_normal = magni.imaging.dictionaries._fastops.dct2(vec, img.shape)

    Perform DCT using the present function:

    >>> from magni.imaging.dictionaries import get_DCT
    >>> matrix = get_DCT(img.shape)
    >>> dct_matrix = matrix.T.dot(vec)

    Check that the two ways produce the same result:

    >>> np.allclose(dct_matrix, dct_normal)
    True

    Compute the overcomplete transform (and back again) and check that the
    resulting image is identical to the original. Notice how this example first
    ensures that the necessary version of SciPy is available:

    >>> from pkg_resources import parse_version
    >>> from scipy import __version__ as _scipy_version
    >>> if parse_version(_scipy_version) >= parse_version('0.16.0'):
    ...     matrix = get_DCT(img.shape, img.shape)
    ...     dct_matrix = matrix.T.dot(vec)
    ...     vec_roundtrip = matrix.dot(dct_matrix)
    ...     np.allclose(vec, vec_roundtrip)
    ... else:
    ...     True
    True

    """

    @_decorate_validation
    def validate_input():
        _levels('shape', (
            _generic(None, 'explicit collection', len_=2),
            _numeric(None, 'integer', range_='[1;inf)')))

        if overcomplete_shape is not None:
            _generic('overcomplete_shape', 'explicit collection', len_=2),
            _numeric(('overcomplete_shape', 0),
                     'integer', range_='[{};inf)'.format(shape[0]))
            _numeric(('overcomplete_shape', 1),
                     'integer', range_='[{};inf)'.format(shape[1]))

    validate_input()

    entries = shape[0] * shape[1]

    if overcomplete_shape is None:
        args = (shape,)
        shape = (entries, entries)
    else:
        if _parse_version(_scipy_version) < _parse_version('0.16.0'):
            raise NotImplementedError(
                'Over-complete DCT requires SciPy >= 0.16.0')

        args = (shape, overcomplete_shape)
        shape = (entries, overcomplete_shape[0] * overcomplete_shape[1])

    return _Matrix(_fastops.idct2, _fastops.dct2, args, shape)


def get_DFT(shape, overcomplete_shape=None):
    """
    Get the DFT fast operation dictionary for the given image shape.

    Parameters
    ----------
    shape : list or tuple
        The shape of the image for which the dictionary is the DFT dictionary.
    overcomplete_shape : list or tuple, optional
        The shape of the (overcomplete) frequency domain for the DFT
        dictionary. The entries must be greater than or equal to the
        corresponding entries in `shape`.

    Returns
    -------
    matrix : magni.utils.matrices.Matrix
        The specified DFT dictionary.

    See Also
    --------
    magni.utils.matrices.Matrix : The matrix emulator class.

    Examples
    --------
    Create a dummy image:

    >>> import numpy as np, magni
    >>> img = np.random.randn(64, 64)
    >>> vec = magni.imaging.mat2vec(img)

    Perform DFT in the ordinary way:

    >>> dft_normal = magni.imaging.dictionaries._fastops.dft2(vec, img.shape)

    Perform DFT using the present function:

    >>> from magni.imaging.dictionaries import get_DFT
    >>> matrix = get_DFT(img.shape)
    >>> dft_matrix = matrix.conj().T.dot(vec)

    Check that the two ways produce the same result:

    >>> np.allclose(dft_matrix, dft_normal)
    True

    Compute the overcomplete transform (and back again):

    >>> matrix = get_DFT(img.shape, img.shape)
    >>> dft_matrix = matrix.conj().T.dot(vec)
    >>> vec_roundtrip = matrix.dot(dft_matrix)

    Check that the twice transformed image is identical to the
    original:

    >>> np.allclose(vec, vec_roundtrip)
    True

    """

    @_decorate_validation
    def validate_input():
        _levels('shape', (
            _generic(None, 'explicit collection', len_=2),
            _numeric(None, 'integer', range_='[1;inf)')))

        if overcomplete_shape is not None:
            _generic('overcomplete_shape', 'explicit collection', len_=2),
            _numeric(('overcomplete_shape', 0),
                     'integer', range_='[{};inf)'.format(shape[0]))
            _numeric(('overcomplete_shape', 1),
                     'integer', range_='[{};inf)'.format(shape[1]))

    validate_input()

    entries = shape[0] * shape[1]

    if overcomplete_shape is None:
        args = (shape,)
        shape = (entries, entries)
    else:
        args = (shape, overcomplete_shape)
        shape = (entries, overcomplete_shape[0] * overcomplete_shape[1])

    return _Matrix(_fastops.idft2, _fastops.dft2, args, shape, is_complex=True)
